import logging
from notifiers import notify


class Notifier:
    """`Notifier` is a library interfacting with different notification providers.

    **Supported providers**

    - email
    - gmail
    - pushover
    - slack
    - telegram
    - twilio

    **Providers not supported yet via specific keywords**

    - gitter
    - join
    - mailgun
    - pagerduty
    - popcornnotify
    - pushbullet
    - simplepush
    - statuspage
    - zulip

    There is a keyword ``Generic Notify`` which can be used
    to call above services, for example.

    .. code-block:: robotframework

        Generic Notify
            provider_name=gitter
            message=Hello from Robot
            token=TOKEN
            room_id=ID_OF_THE_GITTER_ROOM

    Parameters for different providers can be read from the
    **Notifiers** documents (link below).

    Read more at https://notifiers.readthedocs.io/en/latest/


    **About kwargs**

    The `**kwargs` is a term for any extra named parameters, which
    can be included in the same way as already named arguments,
    e.g. ``Notify Email`` could be called with `subject=my email subject`
    which will be passed through `**kwargs`.

    Notifier documentation contains information about all possible
    arguments that different providers support.

    **Robot Framework**

    .. code-block:: robotframework

        &{account}=    Create Dictionary
        ...    host=smtp.office365.com
        ...    username=ACCOUNT_USERNAME
        ...    password=ACCOUNT_PASSWORD
        Notify Email
        ...    to=RECIPIENT_EMAIL
        ...    from_=SENDER_ADDRESS            # passed via kwargs
        ...    subject=Hello from the Robot    # passed via kwargs
        ...    message=Hello from the Robot
        ...    &{account}                      # passed via kwargs

    .. code-block:: python

        notifier = Notifier()
        account = {
            "host": "smtp.office365.com",
            "username": "EMAIL_USERNAME",
            "password": "EMAIL_PASSWORD"
        }
        notifier.email_notify(
            to="RECIPIENT_EMAIL",
            from_="SENDER_EMAIL",
            subject="Hello from the Python Robot",
            message="Hello from the Python RObot",
            **account
        )

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

       *** Settings ***
       Library  RPA.Notifier

       *** Variables ***
       ${SLACK_WEBHOOK}   https://hooks.slack.com/services/WEBHOOKDETAILS
       ${CHANNEL}         notification-channel

       *** Tasks ***
       Lets notify
          Notify Slack   message from robot  channel=${CHANNEL}  webhook_url=${SLACK_WEBHOOK}

    **Python**

    .. code-block:: python

       from RPA.Notifier import Notifier

       library = Notifier()

       slack_attachments = [
          {
             "title": "attachment 1",
             "fallback": "liverpool logo",
             "image_url": "https://upload.wikimedia.org/wikipedia/fi/thumb/c/cd/Liverpool_FC-n_logo.svg/1200px-Liverpool_FC-n_logo.svg.png",
          }
       ]

       library.notify_slack(
          message='message for the Slack',
          channel="notification-channel",
          webhook_url=slack_webhook_url,
          attachments=slack_attachments,
       )
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def notify_pushover(
        self, message: str = None, user: str = None, token: str = None, **kwargs
    ) -> bool:
        """Notify using Pushover provider

        :param message: notification message
        :param user: target user for the notification
        :param token: service token
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not
        """
        arguments = {
            "provider_name": "pushover",
            "message": message,
            "user": user,
            "token": token,
        }
        notify_arguments = {**arguments, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def notify_slack(
        self,
        message: str = None,
        channel: str = None,
        webhook_url: str = None,
        **kwargs,
    ) -> bool:
        """Notify using Slack provider

        :param message: notification message
        :param channel: target channel for the notification
        :param webhook_url: Slack webhook url
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not
        """
        arguments = {
            "provider_name": "slack",
            "message": message,
            "webhook_url": webhook_url,
            "channel": channel,
        }
        notify_arguments = {**arguments, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def notify_telegram(
        self,
        message: str = None,
        chat_id: str = None,
        token: str = None,
        **kwargs,
    ) -> bool:
        """Notify using Telegram provider

        :param message: notification message
        :param chat_id: target chat id for the notification
        :param token: service token
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not
        """
        arguments = {
            "provider_name": "telegram",
            "message": message,
            "chat_id": chat_id,
            "token": token,
        }
        notify_arguments = {**arguments, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def notify_gmail(
        self,
        message: str = None,
        to: str = None,
        username: str = None,
        password: str = None,
        **kwargs,
    ) -> bool:
        """Notify using Gmail provider

        :param message: notification message
        :param to: target of email message
        :param username: GMail account username
        :param password: GMail account password
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not
        """
        arguments = {
            "provider_name": "gmail",
            "message": message,
            "to": to,
            "username": username,
            "password": password,
        }
        notify_arguments = {**arguments, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def notify_email(
        self,
        message: str = None,
        to: str = None,
        username: str = None,
        password: str = None,
        host: str = None,
        port: int = 587,
        tls: bool = True,
        **kwargs,
    ) -> bool:
        """Notify using email provider

        :param message: notification message
        :param to: target of email message
        :param username: email account username
        :param password: email account password
        :param host: email SMTP host name
        :param port: email SMTP host port number
        :param tls: should TLS be used (default True)
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not

        Example.

        .. code:: robotframework

            # Notify with Outlook account
            Notify Email
            ...   message=Message from the Robot
            ...   to=RECIPIENT_EMAIL_ADDRESS
            ...   username=OUTLOOK_USERNAME
            ...   password=OUTLOOK_PASSWORD
            ...   host=smtp.office365.com
            ...   subject=Subject of the Message
        """
        arguments = {
            "provider_name": "email",
            "message": message,
            "to": to,
            "username": username,
            "password": password,
            "host": host,
            "port": port,
            "tls": tls,
        }
        notify_arguments = {**arguments, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def notify_twilio(
        self,
        message: str = None,
        number_from: str = None,
        number_to: str = None,
        account_sid: str = None,
        token: str = None,
        **kwargs,
    ) -> bool:
        """Notify using Twilio provider

        :param message: notification message
        :param number_from: number where the message comes from
        :param number_to: number where the messages goes to
        :param account_sid: Twilio account SID
        :param token: Twilio account token
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not
        """
        arguments = {
            "provider_name": "twilio",
            "message": message,
            "from_": number_from,
            "to": number_to,
            "account_sid": account_sid,
            "auth_token": token,
        }
        notify_arguments = {**arguments, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def generic_notify(self, provider_name: str, **kwargs):
        """Generic keyword to use with any notifier provider.

        :param provider_name: name of the notifier service
        :param kwargs: see library documentation
        :return: True if notification was successful, False if not
        """
        notify_arguments = {"provider_name": provider_name, **kwargs}
        response = notify(**notify_arguments)
        return self._handle_response(response)

    def _handle_response(self, response) -> bool:
        if response.status == "Success":
            self.logger.info("Notify %s resulted in Success", response.provider)
            return True
        else:
            self.logger.error("Notify errors: %s", response.errors)
            return False



from enum import Enum
from typing import Optional, Any, Union

from RPA.core.locators import LocatorType
from RPA.core.helpers import delay
from RPA.core.geometry import Point, Region
from RPA.Desktop.keywords import LibraryContext, keyword


class Action(Enum):
    """Possible mouse click actions."""

    click = 0
    left_click = 0
    double_click = 1
    triple_click = 2
    right_click = 3


def to_action(value):
    """Convert value to Action enum."""
    if isinstance(value, Action):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return Action[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown mouse action: {value}") from err


def to_button(value):
    """Convert value to Button enum."""
    # pylint: disable=C0415
    from pynput_robocorp.mouse import Button

    if isinstance(value, Button):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return Button[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown mouse button: {value}") from err


def to_point(location):
    """Converted resolved location to single point, for clicking."""
    if isinstance(location, Point):
        return location
    elif isinstance(location, Region):
        return location.center
    else:
        raise TypeError(f"Unknown location type: {location}")


class MouseKeywords(LibraryContext):
    """Keywords for sending inputs through an (emulated) mouse."""

    def __init__(self, ctx):
        super().__init__(ctx)
        try:
            # pylint: disable=C0415
            from pynput_robocorp.mouse import Controller

            self._mouse = Controller()
            self._error = None
        except ImportError as exc:
            self._error = exc

    def _move(self, location: Union[Point, Region]) -> None:
        """Move mouse to given location."""
        # TODO: Clamp to screen dimensions?
        point = to_point(location)

        with self.buffer():
            self.logger.info("Moving mouse to (%d, %d)", *point)
            self._mouse.position = point.as_tuple()

    def _click(
        self,
        action: Action = Action.click,
        location: Optional[Union[Point, Region]] = None,
    ) -> None:
        """Perform defined mouse action, and optionally move to given point first."""
        # pylint: disable=C0415
        from pynput_robocorp.mouse import Button

        action = to_action(action)

        if location:
            self._move(location)
            delay(0.05)

        with self.buffer():
            self.logger.info("Performing mouse action: %s", action)
            if action is Action.click:
                self._mouse.click(Button.left)
            elif action is Action.double_click:
                self._mouse.click(Button.left, 2)
            elif action is Action.triple_click:
                self._mouse.click(Button.left, 3)
            elif action is Action.right_click:
                self._mouse.click(Button.right)
            else:
                # TODO: mypy should handle enum exhaustivity validation
                raise ValueError(f"Unsupported action: {action}")

    @keyword
    def click(
        self,
        locator: Optional[LocatorType] = None,
        action: Action = Action.click,
    ) -> None:
        """Click at the element indicated by locator.

        :param locator: Locator for click position
        :param action:  Click action, e.g. right click

        Example:

        .. code-block:: robotframework

            Click
            Click    LoginForm.Button
            Click    coordinates:500,200    triple click
        """
        if self._error:
            raise self._error

        action = to_action(action)

        if locator:
            match = self.ctx.wait_for_element(locator)
            self._click(action, match)
        else:
            self._click(action)

    @keyword
    def click_with_offset(
        self,
        locator: Optional[LocatorType] = None,
        x: int = 0,
        y: int = 0,
        action: Action = Action.click,
    ) -> None:
        """Click at a given pixel offset from the given locator.

        :param locator: Locator for click start position
        :param x:       Click horizontal offset in pixels
        :param y:       Click vertical offset in pixels
        :param action:  Click action, e.g. right click

        Example:

        .. code-block:: robotframework

            Click with offset    Robocorp.Logo    y=400
        """
        if self._error:
            raise self._error

        action = to_action(action)

        if locator:
            position = self.ctx.wait_for_element(locator)
            position = to_point(position)
        else:
            position = self.get_mouse_position()

        position = position.move(int(x), int(y))
        self._click(action, position)

    @keyword
    def get_mouse_position(self) -> Point:
        """Get current mouse position in pixel coordinates.

        Example:

        .. code-block:: robotframework

            ${position}=    Get mouse position
            Log    Current mouse position is ${position.x}, ${position.y}
        """
        if self._error:
            raise self._error

        x, y = self._mouse.position
        return Point(x, y)

    @keyword
    def move_mouse(self, locator: LocatorType) -> None:
        """Move mouse to given coordinates.

        :param locator: Locator for mouse position

        Example:

        .. code-block:: robotframework

            Move mouse    Robocorp.Logo
            Move mouse    offset:0,400
        """
        if self._error:
            raise self._error

        match = self.ctx.wait_for_element(locator)
        self._move(match)

    @keyword
    def press_mouse_button(self, button: Any = "left") -> None:
        """Press down mouse button and keep it pressed."""
        if self._error:
            raise self._error

        button = to_button(button)
        with self.buffer():
            self.logger.info("Pressing down mouse button: %s", button)
            self._mouse.press(button)

    @keyword
    def release_mouse_button(self, button: Any = "left") -> None:
        """Release mouse button that was previously pressed."""
        if self._error:
            raise self._error

        button = to_button(button)
        with self.buffer():
            self.logger.info("Releasing mouse button: %s", button)
            self._mouse.release(button)

    @keyword
    def drag_and_drop(
        self,
        source: LocatorType,
        destination: LocatorType,
        start_delay: float = 2.0,
        end_delay: float = 0.5,
    ) -> None:
        """Drag mouse from source to destination while holding the left mouse button.

        :param source:      Locator for start position
        :param destination: Locator for destination position
        :param start_delay: Delay in seconds after pressing down mouse button
        :param end_delay:   Delay in seconds before releasing mouse button
        """
        if self._error:
            raise self._error

        src = self.ctx.wait_for_element(source)
        dst = self.ctx.wait_for_element(destination)

        src = to_point(src)
        dst = to_point(dst)

        self.logger.info("Dragging from (%d, %d) to (%d, %d)", *src, *dst)

        self._move(src)
        self.press_mouse_button()

        with self.buffer(start_delay):
            self._move(dst)

        with self.buffer(end_delay):
            self.release_mouse_button()



from types import SimpleNamespace
from unittest import mock

import pytest
from RPA.Email.ImapSmtp import ImapSmtp
from RPA.Email.common import counter_duplicate_path
from docx import Document

from . import RESOURCES_DIR


class Resources:
    image = RESOURCES_DIR / "approved.png"
    email = RESOURCES_DIR / "emails" / f"work-item-documentation.eml"


@pytest.fixture
def library():
    with mock.patch("RPA.Email.ImapSmtp.SMTP"), mock.patch(
        "RPA.Email.ImapSmtp.IMAP4_SSL"
    ):
        lib = ImapSmtp(smtp_server="smtp.gmail.com", imap_server="imap.gmail.com")
        yield lib


@pytest.fixture
def creds():
    return SimpleNamespace(
        account="cosmin@robocorp.com",
        password="robocorp-is-cool",
        oauth2_str=(
            "dXNlcj14b2F1dGhAZ21haWwuY29tAWF1dGg9QmVhcmVyIHlhMjkuQUhFUzZaUktlVVF3SDJ4aG"
            "lya3NTVURTeWpnOW9QdVJNTWFsMDVUeTBjZkZJVF91UmZFU0h3AQE="
        ),
    )


@pytest.fixture(
    params=[
        "person1@domain.com",
        "person2@domain.com,person3@domain.com",
    ]
)
def recipients(request):
    return request.param


def test_send_message_all_required_parameters_given(library, recipients):
    status = library.send_message(
        sender="sender@domain.com",
        subject="My test email subject",
        body="body of the message",
        recipients=recipients,
    )
    assert status


def test_send_message_with_no_sender(library, recipients):
    with pytest.raises(TypeError):
        library.send_message(
            subject="My test email subject",
            body="body of the message",
            recipients=recipients,
        )


def test_send_message_with_no_recipients(library):
    with pytest.raises(TypeError):
        library.send_message(
            sender="sender@domain.com",
            subject="My test email subject",
            body="body of the message",
        )


def test_send_message_with_images(library, recipients):
    status = library.send_message(
        sender="sender@domain.com",
        subject="My test email subject",
        body="body of the message<img src='approved.png'/>",
        recipients=recipients,
        html=True,
        images=Resources.image,
    )
    assert status


def test_send_message_with_attachments(library, recipients):
    status = library.send_message(
        sender="sender@domain.com",
        subject="My test email subject",
        body="body of the message",
        recipients=recipients,
        attachments=Resources.image,
    )
    assert status


def test_send_message_with_attachments_and_images(library, recipients):
    status = library.send_message(
        sender="sender@domain.com",
        subject="My test email subject",
        body="body of the message",
        recipients=recipients,
        attachments=Resources.image,
        html=True,
        images=Resources.image,
    )
    assert status


def test_parse_folders_gmail(library):

    folders = [
        b'(\\HasNoChildren) "/" "INBOX"',
        b'(\\HasChildren \\Noselect) "/" "[Gmail]"',
        b'(\\All \\HasNoChildren) "/" "[Gmail]/All Mail"',
    ]

    expected = [
        {"delimiter": "/", "flags": "\\HasNoChildren", "name": "INBOX"},
        {"delimiter": "/", "flags": "\\HasChildren \\Noselect", "name": "[Gmail]"},
        {
            "delimiter": "/",
            "flags": "\\All \\HasNoChildren",
            "name": "[Gmail]/All Mail",
        },
    ]

    result = library._parse_folders(folders)

    assert result == expected


def test_parse_folders_outlook(library):
    folders = [
        b'(\\HasNoChildren) "/" Archive',
        b'(\\HasNoChildren \\Trash) "/" Deleted',
        b'(\\HasNoChildren \\Drafts) "/" Drafts',
    ]
    expected = [
        {"delimiter": "/", "flags": "\\HasNoChildren", "name": "Archive"},
        {"delimiter": "/", "flags": "\\HasNoChildren \\Trash", "name": "Deleted"},
        {"delimiter": "/", "flags": "\\HasNoChildren \\Drafts", "name": "Drafts"},
    ]
    result = library._parse_folders(folders)

    assert result == expected


def test_parse_folders_failed(library, caplog):
    folders = [b"Totally invalid folder_name"]
    expected_log_text = "Cannot parse folder name Totally invalid folder_name"
    result = library._parse_folders(folders)

    assert not result
    assert expected_log_text in caplog.text


def test_email_to_document(tmp_path, library):
    output_source = tmp_path / f"{Resources.email.stem}.docx"
    library.email_to_document(Resources.email, output_source)

    doc = Document(output_source)
    texts = [para.text for para in doc.paragraphs]
    expected_text = (
        "Get attached file from work item to disk. Returns the absolute path to the "
        "created file."
    )
    assert expected_text in texts


def test_basic_authorization(library, creds):
    library.authorize(account=creds.account, password=creds.password)
    library.imap_conn.login.assert_called_once_with(creds.account, creds.password)
    library.smtp_conn.login.assert_called_once_with(creds.account, creds.password)


def test_oauth_authorization(library, creds):
    library.authorize(account=creds.account, password=creds.oauth2_str, is_oauth=True)

    authenticate_call = library.imap_conn.authenticate.call_args[0]
    assert authenticate_call[0] == "XOAUTH2"
    assert b"xoauth@gmail.com" in authenticate_call[1](None)

    auth_call = library.smtp_conn.auth.call_args[0]
    assert auth_call[0] == "XOAUTH2"
    assert "xoauth@gmail.com" in auth_call[1]()


def test_counter_duplicate_path(tmp_path):
    file_path = tmp_path / "my-attachment.txt"
    new_file_path = counter_duplicate_path(file_path)
    assert new_file_path == file_path

    file_path.write_text("some data")
    new_file_path = counter_duplicate_path(file_path)
    assert new_file_path != file_path
    assert new_file_path.name == "my-attachment-2.txt"

    new_file_path.write_text("some data 2")
    newest_file_path = counter_duplicate_path(file_path)
    assert newest_file_path.name == "my-attachment-3.txt"



import base64
import json
import logging
import mimetypes
from typing import List, Optional, Union, Dict

import requests

from RPA.RobotLogListener import RobotLogListener


class Base64AI:
    """Library to support `Base64.ai <https://base64.ai/>`_ service for intelligent
    document processing (IDP).

    Added on **rpaframework** version: 17.0.1

    Service supports identifying fields in the documents, which can be given to the
    service in multiple different file formats and via URL.

    **Robot Framework example usage**

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Base64AI
        Library   RPA.Robocorp.Vault

        *** Tasks ***
        Identify document
            ${secrets}=   Get Secret  base64ai-auth
            Set Authorization  ${secrets}[email-address]   ${secrets}[apikey]
            ${results}=  Scan Document File
            ...   ${CURDIR}${/}invoice.pdf
            ...   model_types=finance/check/usa,finance/invoice/usa
            # Scan response contains list of detected models in the document
            FOR  ${result}  IN  @{results}
                Log To Console  Model: ${result}[model]
                Log To Console  Field keys: ${{','.join($result['fields'].keys())}}
                Log To Console  Fields: ${result}[fields]
                Log To Console  Text (OCR): ${result}[ocr]
            END


    **Python example usage**

    .. code-block:: python

        from RPA.Base64AI import Base64AI
        from RPA.Robocorp.Vault import Vault

        secrets = Vault().get_secret("base64ai-auth")
        baselib = Base64AI()
        baselib.set_authorization(secrets["email-address"], secrets["apikey"])
        result = baselib.scan_document_file(
            "invoice.pdf",
            model_types="finance/invoice,finance/check/usa",
        )
        for r in result:
            print(f"Model: {r['model']}")
            for key, props in r["fields"].items():
                print(f"FIELD {key}: {props['value']}")
            print(f"Text (OCR): {r['ocr']}")
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.base_url = "https://base64.ai/api"
        self._request_headers = {"Content-Type": "application/json"}
        listener = RobotLogListener()
        listener.register_protected_keywords(["RPA.Base64AI.set_authorization"])

    def _get_file_base64_and_mimetype(self, file_path: str):
        with open(file_path, "rb") as image_file:
            encoded_content = base64.b64encode(image_file.read())
        return encoded_content.decode("utf-8"), mimetypes.guess_type(file_path)[0]

    def set_authorization(self, api_email: str, api_key: str) -> None:
        """Set Base64 AI request headers with email and key related to API.

        :param api_email: email address related to the API
        :param api_key: key related to the API

        Robot Framework example:

        .. code-block:: robotframework

            ${secrets}=   Get Secret  base64ai-auth
            Set Authorization    ${secrets}[email-address]    ${secrets}[apikey]

        Python example:

        .. code-block:: python

            secrets = Vault().get_secret("base64ai-auth")
            baselib = Base64AI()
            baselib.set_authorization(secrets["email-address"], secrets["apikey"])
        """
        self._request_headers["Authorization"] = f"ApiKey {api_email}:{api_key}"

    def scan_document_file(
        self,
        file_path: str,
        model_types: Optional[Union[str, List[str]]] = None,
        mock: bool = False,
    ) -> Dict:
        """Scan a document file. Can be given a ``model_types`` to
        specifically target certain models.

        :param file_path: filepath to the file
        :param model_types: single model type or list of model types
        :param mock: set to True to use /mock/scan endpoint instead of /scan
        :return: result of the document scan

        Robot Framework example:

        .. code-block:: robotframework

            ${results}=    Scan Document File
            ...    ${CURDIR}${/}files${/}IMG_8277.jpeg
            ...    model_types=finance/check/usa,finance/invoice
            FOR    ${result}    IN    @{results}
                Log To Console    Model: ${result}[model]
                Log To Console    Fields: ${result}[fields]
                Log To Console    Text (OCR): ${result}[ocr]
            END

        Python example:

        .. code-block:: python

            result = baselib.scan_document_file(
                "./files/Invoice-1120.pdf",
                model_types="finance/invoice,finance/check/usa",
            )
            for r in result:
                print(f"Model: {r['model']}")
                for key, val in r["fields"].items():
                    print(f"{key}: {val['value']}")
                print(f"Text (OCR): {r['ocr']}")
        """
        scan = "mock/scan" if mock else "scan"
        scan_endpoint = f"{self.base_url}/{scan}"
        self.logger.info(f"endpoint {scan_endpoint} is set for scanning")
        base64string, mime = self._get_file_base64_and_mimetype(file_path)
        payload = {"image": f"data:{mime};base64,{base64string}"}
        if model_types:
            req_model_types = (
                model_types if isinstance(model_types, list) else model_types.split(",")
            )
            payload["modelTypes"] = req_model_types

        response = requests.request(
            "POST",
            scan_endpoint,
            headers=self._request_headers,
            data=json.dumps(payload),
        )
        response.raise_for_status()
        return response.json()

    def scan_document_url(
        self,
        url: str,
        model_types: Optional[Union[str, List[str]]] = None,
        mock: bool = False,
    ) -> Dict:
        """Scan a document URL. Can be given a ``model_types`` to
        specifically target certain models.

        :param url: valid url to a file
        :param model_types: single model type or list of model types
        :param mock: set to True to use /mock/scan endpoint instead of /scan
        :return: result of the document scan

        Robot Framework example:

        .. code-block:: robotframework

            ${results}=    Scan Document URL
            ...    https://base64.ai/static/content/features/data-extraction/models//2.png
            FOR    ${result}    IN    @{results}
                Log To Console    Model: ${result}[model]
                Log To Console    Fields: ${result}[fields]
                Log To Console    Text (OCR): ${result}[ocr]
            END

        Python example:

        .. code-block:: python

            result = baselib.scan_document_url(
                "https://base64.ai/static/content/features/data-extraction/models//2.png"
            )
            for r in result:
                print(f"Model: {r['model']}")
                for key, props in r["fields"].items():
                    print(f"FIELD {key}: {props['value']}")
                print(f"Text (OCR): {r['ocr']}")
        """  # noqa: E501
        scan = "mock/scan" if mock else "scan"
        scan_endpoint = f"{self.base_url}/{scan}"
        self.logger.info(f"endpoint {scan_endpoint} is set for scanning")
        payload = {"url": url}
        if model_types:
            req_model_types = (
                model_types if isinstance(model_types, list) else model_types.split(",")
            )
            payload["modelTypes"] = req_model_types

        response = requests.request(
            "POST",
            scan_endpoint,
            headers=self._request_headers,
            data=json.dumps(payload),
        )
        response.raise_for_status()
        self.logger.warning(response.text)
        return response.json()

    def get_user_data(self) -> Dict:
        """Get user data including details on credits used and credits remaining
        for the Base64 service.

        Returned user data contains following keys:

            - givenName
            - familyName
            - email
            - hasWorkEmail
            - companyName
            - numberOfCredits
            - numberOfPages
            - numberOfUploads
            - numberOfCreditsSpentOnDocuments (visible if used)
            - numberOfCreditsSpentOnFaceDetection (visible if used)
            - numberOfCreditsSpentOnFaceRecognition (visible if used)
            - hasActiveAwsContract
            - subscriptionType
            - subscriptionPeriod
            - tags
            - ccEmails
            - status
            - remainingCredits (calculated by the keyword)

        :return: object containing details on the API user

        Robot Framework example:

        .. code-block:: robotframework

            ${userdata}=   Get User Data
            Log To Console  I have still ${userdata}[remainingCredits] credits left

        Python example:

        .. code-block:: python

            userdata = baselib.get_user_data()
            print(f"I have still {userdata['remainingCredits']} credits left")
        """
        response = requests.request(
            "GET",
            f"{self.base_url}/auth/user",
            headers=self._request_headers,
        )
        response.raise_for_status()
        json_response = response.json()
        spent_on_documents = (
            json_response["numberOfCreditsSpentOnDocuments"]
            if "numberOfCreditsSpentOnDocuments" in json_response.keys()
            else 0
        )
        spent_on_face_detection = (
            json_response["numberOfCreditsSpentOnFaceDetection"]
            if "numberOfCreditsSpentOnFaceDetection" in json_response.keys()
            else 0
        )
        spent_on_face_recognition = (
            json_response["numberOfCreditsSpentOnFaceRecognition"]
            if "numberOfCreditsSpentOnFaceRecognition" in json_response.keys()
            else 0
        )
        remainingCredits = (
            json_response["numberOfCredits"]
            - spent_on_documents
            - spent_on_face_detection
            - spent_on_face_recognition
        )
        json_response["remainingCredits"] = remainingCredits
        return json_response



import base64
import os
from itertools import count
from io import BytesIO
from pathlib import Path
from typing import Optional, Union, List, Dict

import mss
from PIL import Image
from robot.api import logger as robot_logger
from robot.running.context import EXECUTION_CONTEXTS
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError

from RPA.core.geometry import Point, Region
from RPA.core.locators import LocatorType
from RPA.Desktop import utils
from RPA.Desktop.keywords import LibraryContext, keyword

if utils.is_windows():
    import ctypes
    from pywinauto import win32structures
    from pywinauto import win32defines
    from pywinauto import win32functions


def _draw_outline(region: Region):
    """Win32-based outline drawing for region."""
    brush_struct = win32structures.LOGBRUSH()
    brush_struct.lbStyle = win32defines.BS_NULL
    brush_struct.lbHatch = win32defines.HS_DIAGCROSS

    brush = win32functions.CreateBrushIndirect(ctypes.byref(brush_struct))
    pen = win32functions.CreatePen(win32defines.PS_SOLID, 2, 0x0000FF)
    dc = win32functions.CreateDC("DISPLAY", None, None, None)

    try:
        win32functions.SelectObject(dc, brush)
        win32functions.SelectObject(dc, pen)
        win32functions.Rectangle(
            dc, region.left, region.top, region.right, region.bottom
        )
    finally:
        win32functions.DeleteObject(brush)
        win32functions.DeleteObject(pen)
        win32functions.DeleteDC(dc)


def _create_unique_path(template: Union[Path, str]) -> Path:
    """Creates a unique path from template with `{index}` placeholder."""
    template = str(template)

    if "{index}" not in template:
        return Path(template)

    for index in count(1):
        path = Path(str(template).format(index=index))
        if not path.is_file():
            return path

    raise RuntimeError("Failed to generate unique path")  # Should not reach here


def _monitor_to_region(monitor: Dict) -> Region:
    """Convert mss monitor to Region instance."""
    return Region.from_size(
        monitor["left"], monitor["top"], monitor["width"], monitor["height"]
    )


def displays() -> List[Region]:
    """Returns list of display regions, without combined virtual display."""
    with mss.mss() as sct:
        return [_monitor_to_region(monitor) for monitor in sct.monitors[1:]]


def grab(region: Optional[Region] = None) -> Image.Image:
    """Take a screenshot of either the full virtual display,
    or a cropped area of the given region.
    """

    with mss.mss() as sct:
        display = _monitor_to_region(sct.monitors[0])

        if region is not None:
            try:
                region = region.clamp(display)
                screenshot = sct.grab(region.as_tuple())
            except ValueError as err:
                raise ValueError("Screenshot region outside display bounds") from err
        else:
            # First monitor is combined virtual display of all monitors
            screenshot = sct.grab(display.as_tuple())

    return Image.frombytes("RGB", screenshot.size, screenshot.bgra, "raw", "BGRX")


def log_image(image: Image.Image, size=1024):
    """Embed image into Robot Framework log."""
    if EXECUTION_CONTEXTS.current is None:
        return

    image = image.copy()
    image.thumbnail((int(size), int(size)), Image.ANTIALIAS)

    buf = BytesIO()
    image.save(buf, format="PNG")
    content = base64.b64encode(buf.getvalue()).decode("utf-8")

    robot_logger.info(
        '<img alt="screenshot" class="rpaframework-desktop-screenshot" '
        f'src="data:image/png;base64,{content}" >',
        html=True,
    )


class ScreenKeywords(LibraryContext):
    """Keywords for reading screen information and content."""

    @keyword
    def take_screenshot(
        self,
        path: Optional[str] = None,
        locator: Optional[LocatorType] = None,
        embed: bool = True,
    ) -> str:
        """Take a screenshot of the whole screen, or an element
        identified by the given locator.

        :param path: Path to screenshot. The string ``{index}`` will be replaced with
            an index number to avoid overwriting previous screenshots.
        :param locator: Element to crop screenshot to
        :param embed: Embed screenshot into Robot Framework log
        """
        if locator is not None:
            element = self.ctx.wait_for_element(locator)
            if not isinstance(element, Region):
                raise ValueError("Locator must resolve to a region")
            image = grab(element)
        else:
            image = grab()

        if path is None:
            try:
                dirname = Path(BuiltIn().get_variable_value("${OUTPUT_DIR}"))
            except RobotNotRunningError:
                dirname = Path.cwd()
            path = dirname / "desktop-screenshot-{index}.png"

        path: Path = _create_unique_path(path).with_suffix(".png")
        os.makedirs(path.parent, exist_ok=True)

        image.save(path)
        self.logger.info("Saved screenshot as '%s'", path)

        if embed:
            log_image(image)

        return str(path)

    @keyword
    def get_display_dimensions(self) -> Region:
        """Returns the dimensions of the current virtual display,
        which is the combined size of all physical monitors.
        """
        with mss.mss() as sct:
            return _monitor_to_region(sct.monitors[0])

    @keyword
    def highlight_elements(self, locator: LocatorType):
        """Draw an outline around all matching elements."""
        if not utils.is_windows():
            raise NotImplementedError("Not supported on non-Windows platforms")

        matches = self.ctx.find_elements(locator)

        for match in matches:
            if isinstance(match, Region):
                _draw_outline(match)
            elif isinstance(match, Point):
                # TODO: Draw a circle instead?
                region = Region(match.x - 5, match.y - 5, match.x + 5, match.y + 5)
                _draw_outline(region)
            else:
                raise TypeError(f"Unknown location type: {match}")

    @keyword
    def define_region(self, left: int, top: int, right: int, bottom: int) -> Region:
        """
        Return a new ``Region`` with the given dimensions.

        :param left: Left edge coordinate.
        :param top: Top edge coordinate.
        :param right: Right edge coordinate.
        :param bottom: Bottom edge coordinate.

        Usage examples:

        .. code-block:: robotframework

            ${region}=  Define Region  10  10  50  30

        .. code-block:: python

            region = desktop.define_region(10, 10, 50, 30)

        """
        return Region(left, top, right, bottom)

    @keyword
    def move_region(self, region: Region, left: int, top: int) -> Region:
        """
        Return a new ``Region`` with an offset from the given region.

        :param region: The region to move.
        :param left: Amount of pixels to move left/right.
        :param top: Amount of pixels to move up/down.

        Usage examples:

        .. code-block:: robotframework

            ${region}=          Find Element  ocr:"Net Assets"
            ${moved_region}=    Move Region  ${region}  500  0

        .. code-block:: python

            region = desktop.find_element('ocr:"Net Assets"')
            moved_region = desktop.move_region(region, 500, 0)

        """
        return region.move(left, top)

    @keyword
    def resize_region(
        self,
        region: Region,
        left: int = 0,
        top: int = 0,
        right: int = 0,
        bottom: int = 0,
    ) -> Region:
        """
        Return a resized new ``Region`` from a given region.

        Extends edges the given amount outward from the center,
        i.e. positive left values move the left edge to the left.

        :param region: The region to resize.
        :param left: Amount of pixels to resize left edge.
        :param top: Amount of pixels to resize top edge.
        :param right: Amount of pixels to resize right edge.
        :param bottom: Amount of pixels to resize bottom edge.

        Usage examples:

        .. code-block:: robotframework

            ${region}=          Find Element  ocr:"Net Assets"
            ${resized_region}=  Resize Region  ${region}  bottom=10

        .. code-block:: python

            region = desktop.find_element('ocr:"Net Assets"')
            resized_region = desktop.resize_region(region, bottom=10)
        """
        return region.resize(left, top, right, bottom)



import base64
import logging
from typing import List, Optional, Union
from pathlib import Path
from O365 import (
    Account,
    MSGraphProtocol,
    FileSystemTokenBackend,
    directory,
    drive as drive_module,
    sharepoint,
)
from O365.utils import Token, BaseTokenBackend
from O365.utils.utils import (  # noqa: F401 pylint: disable=unused-import
    ME_RESOURCE,
    USERS_RESOURCE,
    GROUPS_RESOURCE,
    SITES_RESOURCE,
)
from robot.api.deco import keyword
from RPA.Tables import Table


DEFAULT_REDIRECT_URI = "https://login.microsoftonline.com/common/oauth2/nativeclient"
DEFAULT_TOKEN_PATH = Path("/temp")
DEFAULT_PROTOCOL = MSGraphProtocol()

DriveType = Union[drive_module.Drive, str]
PathType = Union[Path, str]


class MSGraphAuthenticationError(Exception):
    "Error when authentication fails."


class MSGraphDownloadError(Exception):
    "Error when download fails."


class RobocorpVaultTokenBackend(BaseTokenBackend):
    "A simple Token backend that saves to Robocorp vault"


class SharedItem(drive_module.File):
    """Simple class to add support for shared items.
    Inherits File only to bypass checks in the library.
    """  # pylint: disable=super-init-not-called

    def __init__(self, data, url, con):
        self.name = data["name"]
        self.size = data["size"]
        self.con = con
        self.url = url
        self.object_id = None

    def build_url(self, *args):  # pylint: disable=unused-argument
        """Discard arguments since we already have the correct URL."""
        return self.url


class MSGraph:
    """
    The *MSGraph* library wraps the `O365 package`_, giving robots
    the ability to access the Microsoft Graph API programmatically.

    OAuth Configuration
    -------------------

    Graph's API primarily authenticates via the OAuth 2.0 authorization code grant
    flow or OpenID Connect. This library exposes the OAuth 2.0 flow for robots to
    authenticate on behalf of users. A user must complete an initial authentication
    flow with the help of our `OAuth Graph Example Bot`_.

    For best results, `register an app`_ in Azure AD and configure it as so:

    - The type is "Web App".
    - Redirect URI should be ``https://login.microsoftonline.com/common/oauth2/nativeclient``
    - The app should be a multi-tenant app.
    - ``Accounts in any organizational directory`` is checked.
    - Has relevant permissions enabled, check the `Microsoft Graph permissions reference`_
      for a list of permissions available to MS Graph apps.

    .. _O365 package: https://pypi.org/project/O365
    .. _OAuth Graph Example Bot: https://robocorp.com/portal/
    .. _register an app: https://portal.azure.com/#blade/Microsoft_AAD_RegisteredApps/ApplicationsListBlade
    .. _Microsoft Graph permissions reference: https://docs.microsoft.com/en-us/graph/permissions-reference


    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        token: Optional[Token] = None,  # pylint: disable=unused-argument
        refresh_token: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        vault_backend: bool = False,
        vault_secret: Optional[str] = None,
        file_backend_path: Optional[Path] = DEFAULT_TOKEN_PATH,
    ) -> None:
        """When importing the library to Robot Framework, you can set the
        ``client_id`` and ``client_secret``.

        :param client_id: Application client ID.
        :param client_secret: Application client secret.
        :param token: Not implemented. Will support providing the
         authorization token object directly.
        :param refresh_token: Authorization refresh token obtained from
         prior OAuth flows.
        :param redirect_uri: Must be provided with client ID and client
         secret if not using the default.
        :param vault_backend: Not implemented. Will support using the
         Control Room vault directly for authorization token management.
        :param vault_secret: Not implemented. Will support using the
         Control Room vault directly for authorization token management.
        :param file_backend_path: Indicate a path where the tokens will
         be stored within the robot, the default is ``/temp``.
        """
        self.logger = logging.getLogger(__name__)
        # TODO: Implement a `TokenBackend` that uses Robocorp vault,
        #       if implemented, returned refresh tokens are unnecessary.
        if not vault_backend:
            self.token_backend = FileSystemTokenBackend(
                file_backend_path, "auth_token.txt"
            )
        elif vault_backend and not vault_secret:
            raise ValueError(
                "Argument vault_secret cannot be blank if vault_backend set to True."
            )
        else:
            raise NotImplementedError(
                "Robocorp vault token backend not yet implemented."
            )
        if client_id and client_secret:
            self.configure_msgraph_client(
                client_id, client_secret, refresh_token, redirect_uri
            )
        else:
            self.client = None
            self.redirect_uri = redirect_uri or DEFAULT_REDIRECT_URI

    def _require_client(self):
        if self.client is None:
            raise MSGraphAuthenticationError("The MSGraph client is not configured.")

    def _require_authentication(self):
        self._require_client()
        if not self.client.is_authenticated:
            raise MSGraphAuthenticationError(
                "The MS Graph client is not authenticated."
            )

    def _get_refresh_token(self):
        """Returns the refresh token using the backend if that backend
        is not the Vault.
        """
        try:
            if isinstance(self.token_backend, FileSystemTokenBackend):
                return self.token_backend.load_token().get("refresh_token")
            else:
                return None
        except AttributeError:
            return None

    @keyword
    def get_drive_instance(
        self, resource: Optional[str] = None, drive_id: Optional[str] = None
    ) -> drive_module.Drive:
        """Returns the specified drive from the specified resource. If
        either is not provided, defaults are returned.

        :param resource: Name of the resource if not using default.
        :param drive_id: The Drive ID as a string.
        :return: The requested ``Drive`` object.

        .. code-block: robotframework

            *** Tasks ***
            Get default drive
                ${default_drive}=    Get drive instance
                Log    ${default_drive.Name}
        """
        storage = self.client.storage(resource=resource)
        if drive_id:
            return storage.get_drive(drive_id)
        else:
            return storage.get_default_drive()

    def _parse_drive_param(self, drive: DriveType) -> drive_module.Drive:
        """Helper function which parses the drive parameter given
        from a robot.
        """
        if isinstance(drive, drive_module.Drive):
            return drive
        else:
            return self.get_drive_instance(drive)

    @keyword
    def get_folder_instance(
        self,
        drive: Optional[DriveType] = None,
        folder: Optional[Union[drive_module.Folder, str]] = None,
    ) -> drive_module.Folder:
        """Returns a folder object from the provided drive. If a folder
        object is provided, it is reobtained from the API.

        :param drive: A ``Drive`` object or Drive ID as a string. If not
         provided, will use the default drive.
        :param folder: A ``Folder`` object or folder path as a string.

        .. code-block: robotframework

            *** Tasks ***
            Get folder from OneDrive
                ${folder}=    Get folder instance    /path/to/folder
        """
        drive = self._parse_drive_param(drive)
        if isinstance(folder, drive_module.Folder):
            return drive.get_item(folder.object_id)
        elif isinstance(folder, drive_module.DriveItem):
            raise TypeError("The provided DriveItem is not a Folder")
        elif folder in [None, "/", "\\", "root", "ROOT", ""]:
            return drive.get_root_folder()
        else:
            return drive.get_item_by_path(folder)

    @keyword
    def get_file_instance(
        self,
        drive: DriveType,
        file: Union[drive_module.File, str],
    ) -> drive_module.File:
        """Returns a ``File`` object from the provided drive. If a ``File``
        object is provided, it is reobtained from the API.

        :param drive: A ``Drive`` object or Drive ID as a string. If not
         provided, will use the default drive.
        :param file_path: A ``File`` object or path as a string.

        .. code-block: robotframework

            *** Tasks ***
            Get file from OneDrive
                ${folder}=    Get file instance    /path/to/file
        """
        drive = self._parse_drive_param(drive)
        if isinstance(file, str):
            return drive.get_item_by_path(file)
        elif isinstance(file, drive_module.File):
            return drive.get_item(file.object_id)
        else:
            raise TypeError("Target file is not any of the expected types.")

    def _encode_url(self, file_url: str) -> str:
        """Encodes a link to a format supported by the Graph API."""
        base64_bytes = base64.b64encode(bytes(file_url, "utf-8"))
        base64_string = (
            base64_bytes.decode("utf-8")
            .replace("=", "")
            .replace("/", "_")
            .replace("+", "-")
        )
        return "u!{}".format(base64_string)

    def _download_file(
        self,
        file: drive_module.File,
        to_path: Optional[PathType] = None,
        name: Optional[str] = None,
    ) -> Path:
        """Downloads the file and returns the destination path.
        The O365 library returns only a boolean, so it was necessary
        to define the destination path the same way the library does it.
        """
        if not isinstance(file, drive_module.File):
            raise MSGraphDownloadError("Drive item is not a file.")
        if to_path is None:
            to_path = Path()
        else:
            to_path = Path(to_path)
        # TODO: Does this fail when we define a path we want?
        if not to_path.exists():
            raise FileNotFoundError("{} does not exist".format(to_path))
        if name and not Path(name).suffix and file.name:
            name = name + Path(file.name).suffix
        name = name or file.name
        downloaded_file = to_path / name
        success = file.download(to_path=to_path, name=name)
        if not success:
            raise MSGraphDownloadError("Downloading file failed.")
        return downloaded_file

    def _download_folder(
        self,
        folder: drive_module.Folder,
        to_folder: Optional[PathType] = None,
    ) -> Path:
        """Downloads the content of the folder recursively.
        The O365 library returns only a boolean, so it was necessary to define
        the destination path the same way the O365 library does it.
        """
        if not isinstance(folder, drive_module.Folder):
            raise MSGraphDownloadError("Drive item is not a folder.")
        if to_folder is None:
            downloaded_folder = Path() / folder.name
        else:
            downloaded_folder = Path() / to_folder
        # Method download_contents has no return value.
        folder.download_contents(to_folder=downloaded_folder)
        return downloaded_folder

    def _get_sharepoint_drive(
        self, site: sharepoint.Site, drive: Optional[DriveType] = None
    ) -> drive_module.Drive:
        """Returns the specified SharePoint drive if any or the default one if none."""
        drive = self._parse_drive_param(drive)
        if drive:
            return site.get_document_library(drive.object_id)
        else:
            return site.get_default_document_library()

    def _sharepoint_items_into_dict_list(
        self, items_instance: List[sharepoint.SharepointListItem]
    ) -> List[dict]:
        """Turns a list of SharePointListItem into a list of dictionaries.
        This improves the Robot Developer experience, since the inspection
        of attributes isn't easy in Robot Framework.
        """
        items_list = []
        for item in items_instance:
            item_dict = {"object_id": item.object_id}
            item_dict.update(item.fields)
            items_list.append(item_dict)
        return items_list

    @keyword
    def configure_msgraph_client(
        self,
        client_id: str,
        client_secret: str,
        refresh_token: Optional[str] = None,
        redirect_uri: Optional[str] = None,
    ) -> Union[str, None]:
        """Configures the MS Graph client. If a refresh token is
        known, it can be provided to obtain a current user token
        to authenticate with. A new refresh token is returned
        if one is provided.

        :param client_id: Application client ID.
        :param client_secret: Application client secret.
        :param refresh_token: Authorization refresh token obtained from
         prior OAuth flows.
        :param redirect_uri: Must be provided with client ID and client
         secret if not using the default.
        :return: A new refresh token if one was provided, or ``None``.
        """
        credentials = (client_id, client_secret)
        self.client = Account(credentials, token_backend=self.token_backend)
        if redirect_uri is not None:
            self.redirect_uri = redirect_uri
        else:
            self.redirect_uri = DEFAULT_REDIRECT_URI
        if refresh_token:
            return self.refresh_oauth_token(refresh_token)
        return None

    @keyword
    def get_scopes(self, *scopes: str) -> list:
        # pylint: disable=anomalous-backslash-in-string
        """Returns the proper scope definitions based on the
        provided "scope helpers", which are enumerated below.
        You can pass none to get all scopes. Basic is included
        in all other scopes. The provided object can be passed
        to the ``scopes`` parameter when calling
        \`Generate OAuth Authorization URL\`.

        * ``basic``
        * ``mailbox``
        * ``mailbox_shared``
        * ``message_send``
        * ``message_send_shared``
        * ``message_all``
        * ``message_all_shared``
        * ``address_book``
        * ``address_book_shared``
        * ``address_book_all``
        * ``address_book_all_shared``
        * ``calendar``
        * ``calendar_shared``
        * ``calendar_all``
        * ``calendar_shared_all``
        * ``users``
        * ``onedrive``
        * ``onedrive_all``
        * ``sharepoint``
        * ``sharepoint_dl``
        * ``settings_all``
        * ``tasks``
        * ``tasks_all``
        * ``presence``

        :param scopes: Scopes requested. If left empty, all scopes
         will be returned.
        :return: A list of Scopes which can be passed to authorization
         keywords.
        """  # noqa: W605
        if len(scopes) == 0:
            return DEFAULT_PROTOCOL.get_scopes_for(None)
        else:
            scopes_to_get = [s.lower() for s in scopes]
            if "basic" not in scopes_to_get:
                scopes_to_get.append("basic")
            return DEFAULT_PROTOCOL.get_scopes_for(scopes_to_get)

    @keyword
    def generate_oauth_authorization_url(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        scopes: Optional[list] = None,
    ) -> str:
        # pylint: disable=anomalous-backslash-in-string
        """Generates an authorization URL which must be opened
        by the user to complete the OAuth flow. If no scopes
        are provided, the default scope is used which is all
        scopes defined in the \`Get Scopes\` keyword.

        :param client_id: Application client ID.
        :param client_secret: Application client secret.
        :param redirect_uri: Can be provided with client ID and client
         secret if not using the default.
        :param scopes: A list of scopes in the form required by the
         API. Use the helper function \`Get scopes\`.
        :return: The URL the user must follow to complete their portion
         of the OAuth flow.
        """  # noqa: W605
        if redirect_uri is None:
            redirect_uri = (
                self.redirect_uri
                if self.redirect_uri is not None
                else DEFAULT_REDIRECT_URI
            )
        if not self.client:
            self.configure_msgraph_client(
                client_id, client_secret, redirect_uri=redirect_uri
            )
        if scopes is None:
            scopes = self.get_scopes()
        return self.client.connection.get_authorization_url(
            scopes,
            redirect_uri,
        )[0]

    @keyword
    def authorize_and_get_token(self, authorization_url: str) -> str:
        # pylint: disable=anomalous-backslash-in-string
        """Exchanges the OAuth authorization URL obtained from
        \`Generate OAuth Authorization URL\` for an access token. This
        library maintains the user access token for current requests
        and returns the refresh token to be stored in a secure location
        (e.g., the Robocorp Control Room Vault).

        :param authorization_url: The full URL retrieved by the user
         after completing authorization to Microsoft.
        :return: A refresh token which should be stored securely for
         future sessions.
        """  # noqa: W605
        self._require_client()
        if self.client.connection.request_token(
            authorization_url, redirect_uri=self.redirect_uri
        ):
            self.logger.info("Authentication successful.")
            return self._get_refresh_token()
        else:
            raise MSGraphAuthenticationError(
                f"Authentication not successful using '{authorization_url}' as auth URL."  # noqa: E501
            )

    @keyword
    def refresh_oauth_token(self, refresh_token: Optional[str] = None) -> str:
        """Refreshes the user token using the provided ``refresh_token``.
        The user token is retained in the library and a new
        refresh token is returned. If no token is provided, this keyword
        assumes the Robocorp Vault is being used as a backend and attempts
        to refresh it based on that backend.

        :param refresh_token: The refresh token to use to refresh the
         session.
        :return: A refresh token which should be stored securely for
         future sessions.
        """
        self._require_client()
        if refresh_token:
            self.token_backend.token = Token(refresh_token=refresh_token)
            self.token_backend.save_token()
        if self.client.connection.refresh_token():
            self.logger.info("Token successfully refreshed.")
            return self._get_refresh_token()
        else:
            raise MSGraphAuthenticationError("Access token could not be refreshed.")

    @keyword
    def get_me(self) -> directory.User:
        """Returns the MS Graph object representing the currently logged
        in user. A ``User`` object is returned. Properties of the user can
        be accessed like so:

        .. code-block: robotframework

            *** Tasks ***
            Get the me user object
                ${me}=    Get Me
                Log    ${me.full_name}
                Log    ${me.display_name}
                Log    ${me.given_name}
                Log    ${me.surname}
                Log    ${me.full_name}
                Log    ${me.mail}
                Log    ${me.business_phones}
                Log    ${me.mobile_phone}
                Log    ${me.about_me}
                Log    ${me.interests}
                Log    ${me.job_title}
                Log    ${me.object_id}
                Log    ${me.user_principal_name}
        """
        self._require_authentication()
        return self.client.get_current_user()

    @keyword
    def search_for_users(
        self,
        search_string: str,
        search_field: str = "displayName",
        resource: str = USERS_RESOURCE,
    ) -> List[directory.User]:
        # pylint: disable=anomalous-backslash-in-string
        """Returns a list of ``User`` objects from the Active Directory
        based on the provided search string.

        ``User`` objects have additional properties that can be accessed
        with dot-notation, see \`Get Me\` for additional details.

        :param search_string: Text to search for.
        :param search_field: Where to search. Defaults to display name.
        :param resource: Name of the resource if not using default.

        .. code-block: robotframework

            *** Tasks ***
            Search users
                ${users}=    Search For Users    John
        """  # noqa: W605
        self._require_authentication()

        # Get the session if not already defined, which is necessary because we
        # need to modify the headers.
        if self.client.con.session is None:
            self.client.con.session = self.client.con.get_session(load_token=True)

        # It is necessary to pass a specific header to use $search, as the error
        # message instructs: "Request with $search query parameter only works through
        # MSGraph with a special request header: 'ConsistencyLevel: eventual'".
        active_directory = self.client.directory(resource)
        query = active_directory.new_query().search(f"{search_field}:{search_string}")
        self.client.con.session.headers["ConsistencyLevel"] = "eventual"
        users = active_directory.get_users(query=query)
        del self.client.con.session.headers["ConsistencyLevel"]
        return users

    @keyword
    def list_files_in_onedrive_folder(
        self,
        target_folder: Union[drive_module.Folder, str, None] = None,
        include_folders: Optional[bool] = False,
        resource: Optional[str] = None,
        drive_id: Optional[str] = None,
    ) -> List[drive_module.DriveItem]:
        """Returns a list of files from the specified OneDrive folder.

        The files returned are ``DriveItem`` objects and they have additional
        properties that can be accessed with dot-notation.

        :param target_folder: Path of the folder in Onedrive_module.
        :param include_folders: Boolean indicating if should return folders as well.
        :param resource: Name of the resource if not using default.
        :param drive_id: Drive ID if not using default.
        :return: List of DriveItems in the folder.

        .. code-block: robotframework

            *** Tasks ***
            List files
                ${files}=    List Files In Onedrive Folder    /path/to/folder
                FOR    ${file}    IN    @{files}
                    Log    ${file.name}
                    Log    ${file.extension}
                    Log    ${file.description}
                    Log    ${file.created_by}
                    Log    ${file.created}
                    Log    ${file.is_file}
                    Log    ${file.is_image}
                    Log    ${file.is_photo}
                    Log    ${file.is_folder}
                    Log    ${file.size}
                    Log    ${file.web_url}
                END
        """
        self._require_authentication()
        drive_instance = self.get_drive_instance(resource, drive_id)
        folder = self.get_folder_instance(drive_instance, target_folder)
        items = folder.get_items()
        if include_folders:
            return items
        return [item for item in items if not item.is_folder]

    @keyword
    def download_file_from_onedrive(
        self,
        target_file: Union[drive_module.File, str],
        to_path: Optional[PathType] = None,
        name: Optional[str] = None,
        resource: Optional[str] = None,
        drive_id: Optional[str] = None,
    ) -> Path:
        """Downloads a file from Onedrive_module.

        The downloaded file will be saved to a local path.

        :param target_file: ``DriveItem`` or file path of the desired file.
        :param to_path: Destination folder of the downloaded file,
         defaults to the current directory.
        :param name: New name for the downloaded file, with or without extension.
        :param resource: Name of the resource if not using default.
        :param drive_id: Drive ID if not using default.
        :return: Path to the downloaded file.

        .. code-block: robotframework

            *** Tasks ***
            Download file with path
                ${download_path}=    Download File From Onedrive
                ...    /path/to/onedrive/file
                ...    /path/to/local/folder
                ...    Report.pdf

            Download file with object
                ${download_path}=    Download File From Onedrive
                ...    ${drive_item}
                ...    /path/to/local/folder
                ...    Report.pdf
        """
        self._require_authentication()
        drive_instance = self.get_drive_instance(resource, drive_id)
        file_instance = self.get_file_instance(drive_instance, target_file)
        return self._download_file(file_instance, to_path, name)

    @keyword
    def download_folder_from_onedrive(
        self,
        target_folder: Union[drive_module.Folder, str],
        to_path: Optional[PathType] = None,
        resource: Optional[str] = None,
        drive_id: Optional[str] = None,
    ) -> Path:
        """Downloads a folder from OneDrive with all of its contents,
        including subfolders.

        Caution when downloading big folder structures. The downloaded
        folder will be saved to a local path.

        :param target_folder: ``DriveItem`` or path of the desired folder.
        :param to_path: Destination folder where the download will be saved to,
         defaults to the current directory.
        :param resource: Name of the resource if not using default.
        :param drive_id: Drive ID if not using default.
        :return: Path to the downloaded folder.

        .. code-block: robotframework

            *** Tasks ***
            Download folder with path
                ${download_path}=    Download Folder From Onedrive
                ...    /path/to/onedrive/folder
                ...    /path/to/local/folder

            Download folder with object
                ${download_path}=    Download Folder From Onedrive
                ...    ${drive_item}
                ...    /path/to/local/folder
        """
        self._require_authentication()
        drive_instance = self.get_drive_instance(resource, drive_id)
        folder_instance = self.get_folder_instance(drive_instance, target_folder)
        return self._download_folder(folder_instance, to_path)

    @keyword
    def find_onedrive_file(
        self,
        search_string: str,
        target_folder: Optional[Union[drive_module.Folder, str]] = None,
        include_folders: Optional[bool] = False,
        resource: Optional[str] = None,
        drive_id: Optional[str] = None,
    ) -> List[drive_module.DriveItem]:
        # pylint: disable=anomalous-backslash-in-string
        """Returns a list of files found in OneDrive based on the search string.
        If a folder is not specified, the search is done in the entire drive and
        may include items that were shared with the user. It is possible to pass
        \`root\` as the target folder in order to avoid this behavior.

        The files returned are ``DriveItem`` objects and they have additional
        properties that can be accessed with dot-notation, see
        \`List Files In Onedrive Folder\` for details.

        :param search_string: String used to search for file in Onedrive_module.
         Values may be matched across several fields including filename,
         metadata, and file content.
        :param target_folder: Folder where to search for files.
        :param include_folders: Boolean indicating if should return folders as well.
        :param resource: Name of the resource if not using default.
        :param drive_id: Drive ID if not using default.
        :return: List of DriveItems found based on the search string.

        .. code-block: robotframework

            *** Tasks ***
            Find file
                ${files}=    Find Onedrive File    Report.xlsx
        """  # noqa: W605
        self._require_authentication()
        drive_instance = self.get_drive_instance(resource, drive_id)
        if target_folder:
            folder = self.get_folder_instance(drive_instance, target_folder)
            items = folder.search(search_string)
        else:
            items = drive_instance.search(search_string)
        if include_folders:
            return items
        return [item for item in items if not item.is_folder]

    @keyword
    def download_file_from_share_link(
        self,
        share_url: str,
        to_path: Optional[PathType] = None,
        name: Optional[str] = None,
    ) -> Path:
        """Downloads file from the share link.

        The downloaded file will be saved to a local path.

        :param share_url: URL of the shared file
        :param to_path: Destination folder of the downloaded file,
         defaults to the current directory.
        :param name: New name for the downloaded file, with or without extension.

        .. code-block: robotframework

            *** Tasks ***
            Download file
                ${download_path}=    Download File From Share Link
                ...    https://...
                ...    /path/to/local/folder
                ...    Report.pdf
        """
        self._require_authentication()

        # O365 doesn't support getting items from shared links yet
        base_url = self.client.protocol.service_url
        base_url = base_url[:-1] if base_url.endswith("/") else base_url
        encoded_url = self._encode_url(share_url)
        endpoint = "/shares/{id}/driveItem"
        direct_url = "{}{}".format(base_url, endpoint.format(id=encoded_url))

        response = self.client.con.get(direct_url)
        if not response:
            return None

        data = response.json()
        file_instance = SharedItem(data, direct_url, self.client.con)
        return self._download_file(file_instance, to_path, name)

    @keyword
    def upload_file_to_onedrive(
        self,
        file_path: str,
        target_folder: Optional[Union[drive_module.Folder, str]] = None,
        resource: Optional[str] = None,
        drive_id: Optional[str] = None,
    ) -> drive_module.DriveItem:
        # pylint: disable=anomalous-backslash-in-string
        """Uploads a file to the specified OneDrive folder.

        The uploaded file is returned as a ``DriveItem`` object and it has
        additional properties that can be accessed with dot-notation, see
        \`List Files In Onedrive Folder\` for details.

        :param file_path: Path of the local file being uploaded.
        :param target_folder: Path of the folder in Onedrive_module.
        :param resource: Name of the resource if not using default.
        :param drive_id: Drive ID if not using default.

        .. code-block: robotframework

            *** Tasks ***
            Upload file
                ${file}=    Upload File To Onedrive
                ...    /path/to/file.txt
                ...    /path/to/folder
        """  # noqa: W605
        self._require_authentication()
        drive_instance = self.get_drive_instance(resource, drive_id)
        folder = self.get_folder_instance(drive_instance, target_folder)
        return folder.upload_file(item=file_path)

    @keyword
    def get_sharepoint_site(
        self, *args: str, resource: Optional[str] = ""
    ) -> sharepoint.Site:
        """Returns a SharePoint site.

        :param args: The SharePoint site to retrieve, see documentation
         for different methods of defining the site.
        :param resource: Name of the resource if not using default.
        :return: SharePoint Site instance.

        There are multiple ways of retrieving a SharePoint site:

        .. code-block: robotframework

            *** Tasks ***
            Get SharePoint site with host name
                [Documentnation]    Use the host_name e.g.
                ...    'contoso.sharepoint.com' or 'root'.
                Get SharePoint site    contoso.sharepoint.com

            Get SharePoint site with site ID
                [Documentnation]    the site_id is a comma separated string
                ...    of (host_name, site_collection_id, site_id).
                Get SharePoint site
                ...    contoso.sharepoint.com, 1234-1234, 5678-5678

            Get SharePoint site with path to Site
                [Documentnation]    host_name e.g. 'contoso.sharepoint.com'
                ...    and path_to_site is a url path (with a leading slash).
                Get SharePoint site    contoso.sharepoint.com
                ...    /path/to/site

            Get SharePoint site with separate ID args
                [Documentnation]    the site ID broken into separate args
                Get SharePoint site    contoso.sharepoint.com
                ...    1234-1234
                ...    5678-5678

        The return is of type Site and it has additional properties
        that can be accessed with dot-notation. See examples below.

        .. code-block: robotframework

            *** Tasks ***
            Get site
                ${site}=    Get Sharepoint Site    contoso.sharepoint.com
                Log    ${site.name}
                Log    ${site.display_name}
                Log    ${site.description}
                Log    ${site.web_url}
                Log    ${site.object_id}
        """
        self._require_authentication()
        sp = self.client.sharepoint(resource=resource)
        return sp.get_site(*args)

    @keyword
    def get_items_from_sharepoint_list(
        self,
        list_name: str,
        site: sharepoint.Site,
    ) -> Table:
        # pylint: disable=anomalous-backslash-in-string
        """Returns the items on a SharePoint list as a table. The list
        is found by it's display name.

        :param list_name: Display name of the SharePoint list.
        :param site: Site instance obtained from \`Get Sharepoint Site\`.
        :return: Table or list of dicts of the items.

        .. code-block: robotframework

            *** Tasks ***
            Get List
                ${table}=    Get Items From Sharepoint List    My List    ${site}
        """  # noqa: W605
        self._require_authentication()
        sp_list = site.get_list_by_name(list_name)
        sp_items = sp_list.get_items(expand_fields=True)
        items = self._sharepoint_items_into_dict_list(sp_items)

        return Table(items)

    @keyword
    def create_sharepoint_list(
        self,
        list_data: dict,
        site: sharepoint.Site,
    ) -> sharepoint.SharepointList:
        # pylint: disable=anomalous-backslash-in-string
        """Creates a sharepoint list and returns the instance.

        :param list_data: Dictionary with the data for the new list.
        :param site: Site instance obtained from \`Get Sharepoint Site\`.
        :return: SharePoint List that was created.

        ``SharepointList`` objects have additional properties that can be accessed
        with dot-notation, see examples below.

        .. code-block: robotframework

            *** Tasks ***
            Create list
                ${list}=    Create Sharepoint List
                ...    ${list_data}
                ...    ${site}
                Log    ${list.object_id}
                Log    ${list.name}
                Log    ${list.display_name}
                Log    ${list.description}
                Log    ${list.column_name_cw}
                Log    ${list.created_by}
                Log    ${list.created}
                Log    ${list.last_modified_by}
                Log    ${list.modified}
                Log    ${list.web_url}
        """  # noqa: W605
        self._require_authentication()
        return site.create_list(list_data)

    @keyword
    def list_sharepoint_site_drives(
        self, site: sharepoint.Site
    ) -> List[drive_module.Drive]:
        # pylint: disable=anomalous-backslash-in-string
        """Get a list of Drives available in the SharePoint Site.

        :param site: Site instance obtained from \`Get Sharepoint Site\`.
        :return: List of Drives present in the SharePoint Site.

        .. code-block: robotframework

            *** Tasks ***
            List SharePoint drives
                ${drives}    List Sharepoint Site Drives    ${site}
                FOR    ${drive}    IN    @{drives}
                    Log    ${drive.name}
                    Log    ${drive.description}
                    Log    ${drive.owner.display_name}
                    Log    ${drive.web_url}
                    Log    ${drive.object_id}
                END
        """  # noqa: W605
        self._require_authentication()
        return site.list_document_libraries()

    @keyword
    def list_files_in_sharepoint_site_drive(
        self,
        site: sharepoint.Site,
        include_folders: Optional[bool] = False,
        drive_id: Optional[str] = None,
    ) -> List[drive_module.DriveItem]:
        # pylint: disable=anomalous-backslash-in-string
        """List files in the SharePoint Site drive_module.

        If the ``drive_id`` is not set, the default Document Library
        will be used. The ``drive_id`` can be obtained from the keyword
        \`List Sharepoint Site Drives\`.

        The files returned are ``DriveItem`` objects and they have additional
        properties that can be accessed with dot-notation, see
        \`List Files In Onedrive Folder\` for details.

        :param site: Site instance obtained from \`Get Sharepoint Site\`.
        :param include_folders: Boolean indicating if should return folders
         as well.
        :param drive_id: ID of the desired drive_module.
        :return: List of DriveItems present in the Site drive_module.

        .. code-block: robotframework

            *** Tasks ***

            List files in SharePoint drive
                ${files}    List Files In Sharepoint Site Drive    ${site}
        """  # noqa: W605
        self._require_authentication()
        sp_drive = self._get_sharepoint_drive(site, drive_id)
        folder = self.get_folder_instance(sp_drive)
        items = folder.get_items()
        if include_folders:
            return items
        return [item for item in items if not item.is_folder]

    @keyword
    def download_file_from_sharepoint(
        self,
        target_file: Union[drive_module.File, str],
        site: sharepoint.Site,
        to_path: Optional[PathType] = None,
        name: Optional[str] = None,
        drive_id: Optional[str] = None,
    ) -> Path:
        # pylint: disable=anomalous-backslash-in-string
        """Downloads file from SharePoint.

        The downloaded file will be saved to a local folder.

        :param target_file: ``DriveItem`` or file path of the desired file.
        :param site: Site instance obtained from \`Get Sharepoint Site\`.
        :param to_path: Destination folder of the downloaded file,
         defaults to the current directory.
        :param name: New name for the downloaded file, with or without extension.
        :param drive_id: Drive ID if not using default.
        :return: Path to the downloaded file.

        .. code-block: robotframework

            *** Tasks ***
            Download file
                ${download_path}=    Download File From Sharepoint
                ...    /path/to/sharepoint/file
                ...    ${site}
                ...    /path/to/local/folder
                ...    Report.pdf

            Download file with object
                ${download_path}=    Download File From Onedrive
                ...    ${drive_item}
                ...    ${site}
                ...    /path/to/local/folder
                ...    Report.pdf
        """  # noqa: W605
        self._require_authentication()
        sp_drive = self._get_sharepoint_drive(site, drive_id)
        file_instance = self.get_file_instance(sp_drive, target_file)
        return self._download_file(file_instance, to_path, name)



import atexit
import logging
import platform
import struct
from itertools import count
from pathlib import Path
from typing import Any
from contextlib import contextmanager

if platform.system() == "Windows":
    import win32api
    import win32com.client
    import pywintypes
else:
    logging.getLogger(__name__).warning(
        "RPA.Excel.Application library works only on Windows platform"
    )


def _to_unsigned(val):
    return struct.unpack("L", struct.pack("l", val))[0]


@contextmanager
def catch_com_error():
    """Try to convert COM errors to human readable format."""
    try:
        yield
    except pywintypes.com_error as err:  # pylint: disable=no-member
        if err.excepinfo:
            try:
                msg = win32api.FormatMessage(_to_unsigned(err.excepinfo[5]))
            except Exception:  # pylint: disable=broad-except
                msg = err.excepinfo[2]
        else:
            try:
                msg = win32api.FormatMessage(_to_unsigned(err.hresult))
            except Exception:  # pylint: disable=broad-except
                msg = err.strerror
        raise RuntimeError(msg) from err


class Application:
    """`Excel.Application` is a library for controlling an Excel application.

    *Note*. Library works only Windows platform.

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library             RPA.Excel.Application
        Task Setup          Open Application
        Task Teardown       Quit Application

        *** Tasks ***
        Manipulate Excel application
            Open Workbook           workbook.xlsx
            Set Active Worksheet    sheetname=new stuff
            Write To Cells          row=1
            ...                     column=1
            ...                     value=my data
            Save Excel

        Run Excel Macro
            Open Workbook   orders_with_macro.xlsm
            Run Macro       Sheet1.CommandButton1_Click

        Export Workbook as PDF
            Open Workbook           workbook.xlsx
            Export as PDF           workbook.pdf

    **Python**

    .. code-block:: python

        from RPA.Excel.Application import Application

        app = Application()

        app.open_application()
        app.open_workbook('workbook.xlsx')
        app.set_active_worksheet(sheetname='new stuff')
        app.write_to_cells(row=1, column=1, value='new data')
        app.save_excel()
        app.quit_application()
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self, autoexit: bool = True) -> None:
        self.logger = logging.getLogger(__name__)
        self.app = None
        self.workbook = None
        self.worksheet = None

        if platform.system() != "Windows":
            self.logger.warning(
                "Excel application library requires Windows dependencies to work."
            )
        if autoexit:
            atexit.register(self.quit_application)

    def open_application(
        self, visible: bool = False, display_alerts: bool = False
    ) -> None:
        """Open the Excel application.

        :param visible: show window after opening
        :param display_alerts: show alert popups
        """
        with catch_com_error():
            self.app = win32com.client.gencache.EnsureDispatch("Excel.Application")
            self.logger.debug("Opened application: %s", self.app)

            if hasattr(self.app, "Visible"):
                self.app.Visible = visible

            # Show eg. file overwrite warning or not
            if hasattr(self.app, "DisplayAlerts"):
                self.app.DisplayAlerts = display_alerts

    def close_document(self, save_changes: bool = False) -> None:
        """Close the active document (if open)."""
        if not self.workbook:
            return

        with catch_com_error():
            self.workbook.Close(save_changes)

        self.workbook = None
        self.worksheet = None

    def quit_application(self, save_changes: bool = False) -> None:
        """Quit the application."""
        if not self.app:
            return

        self.close_document(save_changes)
        with catch_com_error():
            self.app.Quit()

        self.app = None

    def add_new_workbook(self) -> None:
        """Adds new workbook for Excel application"""
        if not self.app:
            raise ValueError("Excel application is not open")

        with catch_com_error():
            self.workbook = self.app.Workbooks.Add()

    def open_workbook(self, filename: str) -> None:
        """Open Excel by filename

        By default sets active worksheet to sheet number 1

        :param filename: path to filename
        """
        if not self.app:
            self.open_application()

        path = str(Path(filename).resolve())
        self.logger.info("Opening workbook: %s", path)

        with catch_com_error():
            try:
                self.workbook = self.app.Workbooks(path)
            except Exception as exc:  # pylint: disable=broad-except
                self.logger.debug(str(exc))
                self.logger.info("Trying to open workbook by another method")
                self.workbook = self.app.Workbooks.Open(path)

        self.set_active_worksheet(sheetnumber=1)
        self.logger.debug("Current workbook: %s", self.workbook)

    def set_active_worksheet(
        self, sheetname: str = None, sheetnumber: int = None
    ) -> None:
        """Set active worksheet by either its sheet number or name

        :param sheetname: name of Excel sheet, defaults to None
        :param sheetnumber: index of Excel sheet, defaults to None
        """
        if not self.workbook:
            raise ValueError("No workbook open")

        with catch_com_error():
            if sheetnumber:
                self.worksheet = self.workbook.Worksheets(int(sheetnumber))
            elif sheetname:
                self.worksheet = self.workbook.Worksheets(sheetname)

    def add_new_sheet(
        self, sheetname: str, tabname: str = None, create_workbook: bool = True
    ) -> None:
        """Add new worksheet to workbook. Workbook is created by default if
        it does not exist.

        :param sheetname: name for sheet
        :param tabname: name for tab (deprecated)
        :param create_workbook: create workbook if True, defaults to True
        :raises ValueError: error is raised if workbook does not exist and
            `create_workbook` is False
        """
        if tabname:
            sheetname = tabname
            self.logger.warning(
                "Argument 'tabname' is deprecated, "
                "and will be removed in a future version"
            )

        if not self.workbook:
            if not create_workbook:
                raise ValueError("No workbook open")
            self.add_new_workbook()

        self.logger.info("Adding sheet: %s", sheetname)
        with catch_com_error():
            last = self.app.Worksheets(self.app.Worksheets.Count)
            self.worksheet = self.app.Worksheets.Add(After=last)
            self.worksheet.Name = sheetname

    def find_first_available_row(
        self, worksheet: Any = None, row: int = 1, column: int = 1
    ) -> Any:
        """Find first available free row

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: starting row for search, defaults to 1
        :param column: starting column for search, defaults to 1
        :return: row or None
        """
        cell = self.find_first_available_cell(worksheet, row, column)
        # Note: keep return type for backward compability for now
        # return cell[0] if cell else None
        return cell

    def find_first_available_cell(
        self, worksheet: Any = None, row: int = 1, column: int = 1
    ) -> Any:
        """Find first available free cell

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: starting row for search, defaults to 1
        :param column: starting column for search, defaults to 1
        :return: tuple (row, column) or (None, None) if not found
        """
        if not self.workbook:
            raise ValueError("No workbook open")

        if worksheet:
            self.set_active_worksheet(worksheet)

        with catch_com_error():
            for current_row in count(int(row)):
                cell = self.worksheet.Cells(current_row, column)
                if cell.Value is None:
                    return current_row, column

        return None, None

    def write_to_cells(
        self,
        worksheet: Any = None,
        row: int = None,
        column: int = None,
        value: str = None,
        number_format: str = None,
        formula: str = None,
    ) -> None:
        """Write value, number_format and/or formula into cell.

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: target row, defaults to None
        :param column: target row, defaults to None
        :param value: possible value to set, defaults to None
        :param number_format: possible number format to set, defaults to None
        :param formula: possible format to set, defaults to None
        :raises ValueError: if cell is not given
        """
        if not self.workbook:
            raise ValueError("No workbook open")

        if row is None or column is None:
            raise ValueError("No cell was given")

        if worksheet:
            self.set_active_worksheet(worksheet)

        with catch_com_error():
            cell = self.worksheet.Cells(int(row), int(column))

            if number_format:
                cell.NumberFormat = number_format
            if value:
                cell.Value = value
            if formula:
                cell.Formula = formula

    def read_from_cells(
        self,
        worksheet: Any = None,
        row: int = None,
        column: int = None,
    ) -> str:
        """Read value from cell.

        :param worksheet: worksheet to handle, defaults to active worksheet if None
        :param row: target row, defaults to None
        :param column: target row, defaults to None
        :raises ValueError: if cell is not given
        """
        if not self.workbook:
            raise ValueError("No workbook open")

        if row is None or column is None:
            raise ValueError("No cell was given")

        if worksheet:
            self.set_active_worksheet(worksheet)

        with catch_com_error():
            cell = self.worksheet.Cells(int(row), int(column))
            return cell.Value

    def save_excel(self) -> None:
        """Saves Excel file"""
        if not self.workbook:
            raise ValueError("No workbook open")

        with catch_com_error():
            self.workbook.Save()

    def save_excel_as(
        self, filename: str, autofit: bool = False, file_format=None
    ) -> None:
        """Save Excel with name if workbook is open

        :param filename: where to save file
        :param autofit: autofit cell widths if True, defaults to False
        :param file_format: format of file

        **Note:** Changing the file extension for the path does not
        affect the actual format. To use an older format, use
        the ``file_format`` argument with one of the following values:

        https://docs.microsoft.com/en-us/office/vba/api/excel.xlfileformat

        Examples:

        .. code-block:: robotframework

            # Save workbook in modern format
            Save excel as    orders.xlsx

            # Save workbook in Excel 97 format (format from above URL)
            Save excel as    legacy.xls   file_format=${56}
        """
        if not self.workbook:
            # Doesn't raise error for backwards compatibility
            self.logger.warning("No workbook open")
            return

        with catch_com_error():
            if autofit:
                self.worksheet.Rows.AutoFit()
                self.worksheet.Columns.AutoFit()

            path = str(Path(filename).resolve())

            if file_format is not None:
                self.workbook.SaveAs(path, FileFormat=file_format)
            else:
                self.workbook.SaveAs(path)

    def run_macro(self, macro_name: str, *args: Any):
        """Run Excel macro with given name

        :param macro_name: macro to run
        :param args: arguments to pass to macro
        """
        if not self.app:
            raise ValueError("Excel application is not open")

        if not self.workbook:
            raise ValueError("No workbook open")

        with catch_com_error():
            self.app.Application.Run(f"'{self.workbook.Name}'!{macro_name}", *args)

    def export_as_pdf(self, pdf_filename: str, excel_filename: str = None):
        """Export Excel as PDF file

        If Excel filename is not given, the currently open workbook
        will be exported as PDF.

        :param pdf_filename: PDF filename to save
        :param excel_filename: Excel filename to open
        """
        if excel_filename:
            self.open_workbook(excel_filename)
        else:
            if not self.workbook:
                raise ValueError("No workbook open. Can't export PDF.")
        with catch_com_error():
            path = str(Path(pdf_filename).resolve())
            self.workbook.ExportAsFixedFormat(0, path)



from enum import Enum
import logging
from typing import Optional
from pyotp import HOTP, TOTP
from robot.api.deco import keyword
from RPA.Robocorp.Vault import Vault


class OTPMode(Enum):
    """Enumeration for type of TOP to use"""

    TIME = "TIME"
    COUNTER = "COUNTER"


class TOTPNotSetError(Exception):
    "Error when TOTP (Time-based One-Time Password) has not been set"


class HOTPNotSetError(Exception):
    "Error when HOTP (HMAC One-Time Password) has not been set"


TOTP_NOT_SET_ERROR_MSG = """TOTP (Time-based One-Time Password) can be set in library initialization, with
`Use MFA Secret From Vault` keyword or with `Set Time Based OTP` keyword."""

HOTP_NOT_SET_ERROR_MSG = """HOTP (HMAC One-Time Password) can be set in library initialization, with
`Use MFA Secret From Vault` keyword or with `Set Counter Based OTP` keyword."""


class MFA:
    """*RPA.MFA* is a library for generating one-time passwords (OTP).

    Added on **rpaframework** version: 16.0.0

    Based on the `pyotp <https://pypi.org/project/pyotp/>`_ package.

    In the below example the **mfa** secret we are reading from the Robocorp
    Vault is the passcode generated by the Authenticator service. The passcode
    value is stored into the Vault with key **otpsecret**.

    Passcode is typically a long string (16-32 characters), which is provided
    in a form of QR image, but it can be obtained by requesting access to a string.

    Note that same code can be used to add a mobile phone as a authentication
    device at the same as the same code is added into the Vault.

    **Robot framework example usage:**

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Robocorp.Vault
        Library     RPA.MFA


        *** Tasks ***
        Generate time based code
            ${secrets}=    Get Secret   mfa
            ${code}=    Get Time Based OTP    ${secrets}[otpsecret]


    **Python example usage**

    .. code-block:: python

        from RPA.Robocorp.Vault import Vault
        from RPA.MFA import MFA


        def main():
            secrets = Vault().get_secret("mfa")
            code = MFA().get_time_based_otp(secrets["otpsecret"])
    """

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(
        self,
        vault_name: Optional[str] = None,
        vault_key: Optional[str] = None,
        mode: Optional[OTPMode] = OTPMode.TIME,
    ):
        self.logger = logging.getLogger(__name__)
        self._hotp = None
        self._totp = None
        if vault_name and vault_key:
            self.use_mfa_secret_from_vault(vault_name, vault_key, mode)

    @keyword
    def use_mfa_secret_from_vault(
        self, vault_name: str, vault_key: str, mode: OTPMode = OTPMode.TIME
    ):
        """Set `time` or `counter` based OTP with passcode stored in
        the Robocorp Vault named with `vault_name` under key of `vault_key`.

        :param vault_name: name of the vault storing the passcode
        :param vault_key: name of the vault key storing the passcode value
        """
        secrets = Vault().get_secret(vault_name)
        if mode == OTPMode.TIME:
            self.set_time_based_otp(secrets[vault_key])
        elif mode == OTPMode.COUNTER:
            self.set_counter_based_otp(secrets[vault_key])

    @keyword
    def set_time_based_otp(self, otp_passcode: str):
        """Set `time` based OTP with passcode.

        :param otp_passcode: the passcode provided by the Authenticator
        """
        self._totp = TOTP(otp_passcode)

    @keyword
    def set_counter_based_otp(self, otp_passcode: str):
        """Set `counter` based OTP with passcode.

        :param otp_passcode: the passcode provided by the Authenticator
        """
        self._hotp = HOTP(otp_passcode)

    @keyword
    def get_time_based_otp(self, otp_passcode: Optional[str] = None):
        """Get `time` based one time password using separately set
        passcode or by parameter `otp_passcode`.

        :param otp_passcode: the passcode provided by the Authenticator
        """
        if otp_passcode:
            self.set_time_based_otp(otp_passcode)
        if not self._totp:
            raise TOTPNotSetError(TOTP_NOT_SET_ERROR_MSG)
        return self._totp.now()

    @keyword
    def get_counter_based_otp(
        self,
        counter: int,
        otp_passcode: Optional[str] = None,
    ):
        """Get `counter` based one time password using separately set
        passcode or by parameter `otp_passcode`. The counter index is
        given by the `counter` parameter.

        :param counter: the index of the counter
        :param otp_passcode: the passcode provided by the Authenticator
        """
        if otp_passcode:
            self.set_counter_based_otp(otp_passcode)
        if not self._hotp:
            raise HOTPNotSetError(HOTP_NOT_SET_ERROR_MSG)
        return self._hotp.at(counter)



import base64
import logging
import os
import re
import time

# email package declares these properties in the __all__ definition but
# pylint ignores that
from email import encoders, message_from_bytes  # pylint: disable=E0611
from email.charset import QP, add_charset  # pylint: disable=E0611
from email.generator import Generator  # pylint: disable=E0611
from email.header import Header, decode_header, make_header  # pylint: disable=E0611
from email.message import Message  # pylint: disable=E0611
from email.mime.base import MIMEBase  # pylint: disable=E0611
from email.mime.image import MIMEImage  # pylint: disable=E0611
from email.mime.multipart import MIMEMultipart  # pylint: disable=E0611
from email.mime.text import MIMEText  # pylint: disable=E0611
from enum import Enum
from functools import wraps
from imaplib import IMAP4_SSL
from io import StringIO
from pathlib import Path
from smtplib import (
    SMTP,
    SMTP_SSL,
    SMTPConnectError,
    SMTPNotSupportedError,
    SMTPServerDisconnected,
    ssl,
)
from typing import Any, BinaryIO, List, Optional, Tuple, Union

from htmldocx import HtmlToDocx
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from RPA.Email.common import counter_duplicate_path
from RPA.RobotLogListener import RobotLogListener

FilePath = Union[str, Path]


class AttachmentPosition(Enum):
    """Possible attachment positions in the message content."""

    TOP = 1  # Default
    BOTTOM = 2


def to_attachment_position(value):
    """Convert value to AttachmentPosition enum."""
    if isinstance(value, AttachmentPosition):
        return value

    sanitized = str(value).upper().strip().replace(" ", "_")
    try:
        return AttachmentPosition[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown AttachmentPosition: {value}") from err


class Action(Enum):
    """Possible email actions."""

    msg_copy = 1
    msg_delete = 2
    msg_list = 3
    msg_flag = 4
    msg_unflag = 5
    msg_read = 6
    msg_unread = 7
    msg_save = 8
    msg_attachment_save = 9
    msg_move = 10
    msg_trash = 11
    glabel_add = 20  # Add GMail label
    glabel_remove = 21  # Remove GMail label


def to_action(value):
    """Convert value to Action enum."""
    if isinstance(value, Action):
        return value

    sanitized = str(value).lower().strip().replace(" ", "_")
    try:
        return Action[sanitized]
    except KeyError as err:
        raise ValueError(f"Unknown email action: {value}") from err


def get_part_filename(msg: Message) -> Optional[str]:
    filename = msg.get_filename()
    if not filename:
        return None

    decoded = decode_header(filename)
    if decoded[0][1] is not None:
        filename = decoded[0][0].decode(decoded[0][1])
    return filename.replace("\r", "").replace("\n", "")


IMAGE_FORMATS = ["jpg", "jpeg", "bmp", "png", "gif"]
FLAG_DELETED = "\\Deleted"
FLAG_SEEN = "\\Seen"
FLAG_FLAGGED = "\\Flagged"
FLAG_TRASH = "\\Trash"

try:
    BuiltIn().import_library("RPA.RobotLogListener")
except RobotNotRunningError:
    pass


def imap_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].imap_conn is None:
            raise ValueError("Requires authorized IMAP connection")
        return f(*args, **kwargs)

    return wrapper


def smtp_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        if args[0].smtp_conn is None:
            raise ValueError("Requires authorized SMTP connection")
        return f(*args, **kwargs)

    return wrapper


class ImapSmtp:
    """`ImapSmtp` is a library for sending, reading, and deleting emails.
    `ImapSmtp` is interfacing with SMTP and IMAP protocols.

    ***About criteria argument***

    Various keywords like ``List Messages`` and ``Move Messages`` have keyword
    argument called ``criterion`` which can be used to filter emails according
    to given criteria.

    Syntax needs to according to specification and more information about that
    can be read from https://robocorp.com/docs/development-guide/email/sending-emails-with-gmail-smtp#listing-email-messages-by-criteria

    **Troubleshooting**

    - Authentication error with Gmail - "Application-specific password required"
        see. https://support.google.com/mail/answer/185833?hl=en
    - More secure apps (XOAUTH2 protocol): Check
        `example-oauth-email <https://github.com/robocorp/example-oauth-email>`_

    **Examples**

    **Robot Framework**

    It is highly recommended to secure your passwords and take care
    that they are not stored in version control by mistake.
    See ``RPA.Robocorp.Vault`` to see how to store secrets in
    Robocorp Vault.

    When sending HTML content with IMG tags, the ``src`` filenames must match
    the base image name given with the ``images`` parameter.

    .. code-block:: robotframework

        *** Settings ***
        Library     RPA.Email.ImapSmtp   smtp_server=smtp.gmail.com  smtp_port=587
        Task Setup  Authorize  account=${GMAIL_ACCOUNT}  password=${GMAIL_PASSWORD}

        *** Variables ***
        ${GMAIL_ACCOUNT}        ACCOUNT_NAME
        ${GMAIL_PASSWORD}       ACCOUNT_PASSWORD
        ${RECIPIENT_ADDRESS}    RECIPIENT
        ${BODY_IMG1}            ${IMAGEDIR}${/}approved.png
        ${BODY_IMG2}            ${IMAGEDIR}${/}invoice.png
        ${EMAIL_BODY}     <h1>Heading</h1><p>Status: <img src='approved.png' alt='approved image'/></p>
        ...               <p>INVOICE: <img src='invoice.png' alt='invoice image'/></p>

        *** Tasks ***
        Sending email
            Send Message  sender=${GMAIL_ACCOUNT}
            ...           recipients=${RECIPIENT_ADDRESS}
            ...           subject=Message from RPA Robot
            ...           body=RPA Robot message body

        Sending HTML Email With Image
            [Documentation]     Sending email with HTML content and attachment
            Send Message
            ...                 sender=${GMAIL_ACCOUNT}
            ...                 recipients=${RECIPIENT_ADDRESS}
            ...                 subject=HTML email with body images (2) plus one attachment
            ...                 body=${EMAIL_BODY}
            ...                 html=${TRUE}
            ...                 images=${BODY_IMG1}, ${BODY_IMG2}
            ...                 attachments=example.png

    **Python**

    .. code-block:: python

        from RPA.Email.ImapSmtp import ImapSmtp

        gmail_account = "ACCOUNT_NAME"
        gmail_password = "ACCOUNT_PASSWORD"
        sender = gmail_account

        mail = ImapSmtp(smtp_server="smtp.gmail.com", smtp_port=587)
        mail.authorize(account=gmail_account, password=gmail_password)
        mail.send_message(
            sender=gmail_account,
            recipients="RECIPIENT",
            subject="Message from RPA Python",
            body="RPA Python message body",
        )
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(
        self,
        smtp_server: str = None,
        smtp_port: int = 587,
        imap_server: str = None,
        imap_port: int = 993,
        account: str = None,
        password: str = None,
        encoding: str = "utf-8",
    ) -> None:
        listener = RobotLogListener()
        listener.register_protected_keywords(
            ["RPA.Email.ImapSmtp.authorize", "RPA.Email.ImapSmtp.set_credentials"]
        )

        self.logger = logging.getLogger(__name__)
        self.smtp_server = smtp_server
        self.imap_server = imap_server if imap_server else smtp_server
        self.smtp_port = int(smtp_port)
        self.imap_port = int(imap_port)
        self.set_credentials(account, password)
        self.smtp_conn = None
        self.imap_conn = None
        self.selected_folder = None
        self.encoding = encoding

    def __del__(self) -> None:
        if self.smtp_conn:
            try:
                self.smtp_conn.quit()
            except SMTPServerDisconnected:
                self.logger.debug("Was already disconnected")
            finally:
                self.smtp_conn = None

        if self.imap_conn:
            try:
                self.imap_conn.close()
                self.imap_conn.logout()
            except Exception:  # pylint: disable=W0703
                pass
            self.imap_conn = None

    def set_credentials(self, account: str = None, password: str = None) -> None:
        """Set credentials

        :param account: user account as string, defaults to None
        :param password: user password as string, defaults to None

        Example:

        .. code-block:: robotframework

            Set Credentials   ${username}   ${password}
            Authorize
        """

        self.account = account
        self.password = password

    def authorize_smtp(
        self,
        account: str = None,
        password: str = None,
        smtp_server: str = None,
        smtp_port: int = None,
        is_oauth: bool = False,
    ) -> None:
        """Authorize to SMTP server.

        :param account: SMTP account name, defaults to None
        :param password: SMTP account password, defaults to None
        :param smtp_server: SMTP server address, defaults to None
        :param smtp_port: SMTP server port, defaults to None (587 for SMTP)
        :param is_oauth: Use XOAUTH2 protocol with a base64 encoded OAuth2 string as
            `password`

        Can be called without giving any parameters if library
        has been initialized with necessary information and/or
        keyword ``Set Credentials`` has been called.

        Example:

        .. code-block:: robotframework

            Authorize SMTP    ${username}   ${password}  smtp.gmail.com  587
        """
        account = account or self.account
        password = password or self.password
        smtp_server = smtp_server or self.smtp_server
        smtp_port = smtp_port or self.smtp_port
        if smtp_server:
            try:
                self.smtp_conn = SMTP(smtp_server, smtp_port)
                self.send_smtp_hello()
                try:
                    self.smtp_conn.starttls()
                except SMTPNotSupportedError:
                    self.logger.warning("TLS not supported by the server")
            except SMTPConnectError:
                context = ssl.create_default_context()
                self.smtp_conn = SMTP_SSL(smtp_server, smtp_port, context=context)
            if account and password:
                if is_oauth:
                    self.smtp_conn.auth(
                        "XOAUTH2", lambda: base64.b64decode(password.encode()).decode()
                    )
                else:
                    self.smtp_conn.login(account, password)
        else:
            self.logger.warning("SMTP server address is needed for authentication")
        if self.smtp_conn is None:
            self.logger.warning("Not able to establish SMTP connection")

    def authorize_imap(
        self,
        account: str = None,
        password: str = None,
        imap_server: str = None,
        imap_port: int = None,
        is_oauth: bool = False,
    ) -> None:
        """Authorize to IMAP server.

        :param account: IMAP account name, defaults to None
        :param password: IMAP account password, defaults to None
        :param imap_server: IMAP server address, defaults to None
        :param imap_port: IMAP server port, defaults to None
        :param is_oauth: Use XOAUTH2 protocol with a base64 encoded OAuth2 string as
            `password`

        Can be called without giving any parameters if library
        has been initialized with necessary information and/or
        keyword ``Set Credentials`` has been called.

        Example:

        .. code-block:: robotframework

            Authorize IMAP    ${username}   ${password}  imap.gmail.com  993
        """
        if account is None and password is None:
            account = self.account
            password = self.password
        if imap_server is None:
            imap_server = self.imap_server
        if imap_port is None:
            imap_port = self.imap_port
        else:
            imap_port = int(imap_port)
        if imap_server and account and password:
            self.imap_conn = IMAP4_SSL(imap_server, imap_port)
            if is_oauth:
                self.imap_conn.authenticate(
                    "XOAUTH2", lambda _: base64.b64decode(password.encode())
                )
            else:
                self.imap_conn.login(account, password)
            self.imap_conn.select("INBOX")
        else:
            self.logger.warning(
                "Server address, account and password are needed for "
                "authentication with IMAP"
            )
        if self.imap_conn is None:
            self.logger.warning("Not able to establish IMAP connection")

    def authorize(
        self,
        account: str = None,
        password: str = None,
        smtp_server: str = None,
        imap_server: str = None,
        smtp_port: int = None,
        imap_port: int = None,
        is_oauth: bool = False,
    ) -> None:
        # pylint: disable=C0301
        """Authorize user to SMTP and IMAP servers.

        :param account: user account as string, defaults to None
        :param password: user password as string, defaults to None
        :param smtp_server: SMTP server address, defaults to None
        :param imap_server: IMAP server address, defaults to None
        :param smtp_port: SMTP server port, defaults to None (587 for SMTP)
        :param imap_port: IMAP server port, defaults to None
        :param is_oauth: Use XOAUTH2 protocol with a base64 encoded OAuth2 string as
            `password`

        Will use separately set credentials or those given in keyword call.

        Example:

        .. code-block:: robotframework

            Authorize    ${username}   ${password}  smtp_server=smtp.gmail.com  smtp_port=587
        """  # noqa: E501
        self.authorize_smtp(
            account, password, smtp_server, smtp_port, is_oauth=is_oauth
        )
        self.authorize_imap(
            account, password, imap_server, imap_port, is_oauth=is_oauth
        )

    @smtp_connection
    def send_smtp_hello(self) -> None:
        """Send hello message to SMTP server.

        Required step when creating SMTP connection.
        """
        if self.smtp_conn:
            self.smtp_conn.ehlo()

    def send_message(
        self,
        sender: str,
        recipients: Union[List[str], str],
        subject: str = "",
        body: str = "",
        attachments: Optional[Union[List[str], str]] = None,
        html: bool = False,
        images: Optional[Union[List[str], str]] = None,
        cc: Optional[Union[List[str], str]] = None,
        bcc: Optional[Union[List[str], str]] = None,
        attachment_position: Optional[AttachmentPosition] = AttachmentPosition.TOP,
    ) -> bool:
        """Send SMTP email

        :param sender: who is sending, ie. 'from'
        :param recipients: who is receiving, ie. 'to'
        :param subject: mail subject field
        :param body: mail body content
        :param attachments: list of filepaths to attach
        :param html: if message content is in HTML, default `False`
        :param images: list of filepaths for inline images
        :param cc: list of email addresses for email 'cc' field
        :param bcc: list of email addresses for email 'bcc' field
        :param attachment_position: content position for attachment, default `top`

        Valid sender values:

        - First Lastname <address@domain>
        - address@domain

        Example:

        .. code-block:: robotframework

            Send Message  sender@domain.com  recipient@domain.com
            ...           cc=need_to_know@domain.com
            ...           bcc=hidden_copy@domain.com
            ...           subject=Greetings Software Robot Developer
            ...           body=${email_body}
            ...           attachments=${CURDIR}${/}report.pdf

            # Fixing attachments to the bottom of the content
            Send Message  sender@domain.com  recipient@domain.com
            ...           subject=Greetings Software Robot Developer
            ...           body=${email_body}
            ...           attachments=${CURDIR}${/}report.pdf
            ...           attachment_position=bottom
        """
        evaluated_attachment_position = to_attachment_position(attachment_position)
        add_charset(self.encoding, QP, QP, self.encoding)
        to, attachments, images = self._handle_message_parameters(
            recipients, attachments, images
        )
        msg = MIMEMultipart()

        if evaluated_attachment_position == AttachmentPosition.TOP:
            self._add_attachments_to_msg(attachments, msg)

        sender = sender.encode("idna").decode("ascii")
        msg_to = ",".join(to).encode("idna").decode("ascii")
        msg["From"] = sender
        msg["To"] = msg_to
        msg["Subject"] = Header(subject, self.encoding)
        recipients = to if isinstance(to, list) else [to]
        if cc:
            msg["Cc"] = ",".join(cc) if isinstance(cc, list) else cc
            recipients += cc if isinstance(cc, list) else cc.split(",")
        if bcc:
            recipients += bcc if isinstance(bcc, list) else bcc.split(",")

        self._add_message_content(html, images, body, msg)

        if evaluated_attachment_position == AttachmentPosition.BOTTOM:
            self._add_attachments_to_msg(attachments, msg)

        # Create a generator and flatten message object to 'file’
        str_io = StringIO()
        g = Generator(str_io, False)
        g.flatten(msg)
        try:
            if self.smtp_conn is None:
                self.authorize_smtp()
            self.smtp_conn.sendmail(sender, recipients, str_io.getvalue())
        except Exception as err:
            raise ValueError(f"Send Message failed: {err}") from err
        return True

    def _add_message_content(self, html, images, body, msg):
        if html:
            for im in images:
                im = im.strip()
                imname = Path(im).name
                body = body.replace(str(imname), f"cid:{imname}")
                with open(im, "rb") as f:
                    img = MIMEImage(f.read())
                    img.add_header("Content-ID", f"<{imname}>")
                    msg.attach(img)
            htmlpart = MIMEText(body, "html", self.encoding)
            msg.attach(htmlpart)
        else:
            textpart = MIMEText(body, "plain", self.encoding)
            msg.attach(textpart)
            for im in images:
                im = im.strip()
                imname = Path(im).name
                with open(im, "rb") as f:
                    img = MIMEImage(f.read())
                    msg.add_header(
                        "Content-Disposition",
                        f"inline; filename= {imname}",
                    )
                    msg.attach(img)

    def _handle_message_parameters(self, recipients, attachments, images):
        if attachments is None:
            attachments = []
        if images is None:
            images = []
        if not isinstance(recipients, list):
            recipients = recipients.split(",")
        if not isinstance(attachments, list):
            attachments = str(attachments).split(",")
        if not isinstance(images, list):
            images = str(images).split(",")
        return recipients, attachments, images

    def _add_attachments_to_msg(self, attachments: list = None, msg=None):
        if len(attachments) > 0:
            for filename in attachments:
                if os.path.dirname(filename) == "":
                    filename = str(Path.cwd() / filename)
                with open(filename, "rb") as attachment:
                    _, ext = filename.lower().rsplit(".", 1)
                    if ext in IMAGE_FORMATS:
                        # image attachment
                        part = MIMEImage(
                            attachment.read(),
                            name=Path(filename).name,
                            _subtype=ext,
                        )
                    else:
                        # attach other filetypes
                        part = MIMEBase("application", "octet-stream")
                        part.set_payload(attachment.read())
                        encoders.encode_base64(part)
                    part.add_header(
                        "Content-Disposition",
                        "attachment",
                        filename=Path(filename).name,
                    )
                    msg.attach(part)

    def get_decoded_email_body(
        self, message, html_first: bool = False
    ) -> Tuple[str, bool]:
        """Decodes email body and extracts its text/html content.

        Automatically detects character set if the header is not set.

        :param message: Raw 7-bit message body input e.g. from `imaplib`. Double
            encoded in quoted-printable and latin-1
        :param html_first: Prioritize html extraction over text when this is True
        :returns: Message body as unicode string and a boolean telling if the message
            has attachments
        """
        if not message.is_multipart():
            content_charset = message.get_content_charset()
            text = str(
                message.get_payload(decode=True),
                content_charset or self.encoding,
                "ignore",
            )
            return text.strip(), False

        text = html = None
        has_attachments = False

        for part in message.walk():
            if not part:
                continue
            content_filename = get_part_filename(part)
            if content_filename:
                has_attachments = True
                continue
            content_type = "text/plain"
            _data = ""
            content_charset = part.get_content_charset()
            if content_charset:
                content_type = part.get_content_type()
            payload = part.get_payload(decode=True)
            if payload:
                _data = str(payload, str(content_charset), "ignore")

            if content_type == "text/plain":
                text = _data
            elif content_type == "text/html":
                html = _data

        if html_first:
            data = html or text
        else:
            data = text or html
        data = data.strip() if data else ""
        return data, has_attachments

    @imap_connection
    def _do_actions_on_messages(
        self,
        criterion: Union[str, dict],
        actions: list,
        labels: str = None,
        source_folder: str = None,
        target_folder: str = None,
        limit: int = None,
        overwrite: bool = False,
        readonly: bool = False,
        prefix: str = "",
    ) -> list:
        selected_folder = source_folder or self.selected_folder
        folders = self.get_folder_list(subdirectory=selected_folder)
        result = {"actions_done": 0, "message_count": 0, "ids": [], "uids": {}}

        if source_folder:
            for folder in folders:
                if "Noselect" in folder["flags"]:
                    continue
                folder_name = folder["name"]
                self.select_folder(folder_name, readonly)
                self._search_message(criterion, actions, limit, result)
        else:
            self._search_message(criterion, actions, limit, result)

        if len(result["uids"].keys()) == 0:
            self.logger.warning(
                "Can't find any messages matching criterion '%s' in source folder '%s'",
                criterion,
                source_folder or "INBOX",
            )
            return result
        if limit is None or len(result["uids"].keys()) <= limit:
            for mail in result["uids"].items():
                status = self._perform_actions(
                    actions,
                    mail,
                    labels=labels,
                    target_folder=target_folder,
                    overwrite=overwrite,
                    prefix=prefix,
                )
                if status:
                    result["actions_done"] += 1
        else:
            self.logger.warning(
                "Some emails were not processed because limit was reached"
            )
        return result

    def _search_message(self, criterion, actions, limit, result):
        search_encoding = None
        search_command = None
        literal_search = criterion.startswith("literal:")
        gmail_search = criterion.startswith("gmail:")

        if literal_search:
            search_encoding = "utf-8"
            search_term = criterion.replace("literal:", "")
            search_command, search_literal = search_term.split(" ", 1)

            self.imap_conn.literal = b"%s" % search_literal.encode("utf-8")
        elif gmail_search:
            search_encoding = "utf-8"
            search_command = "X-GM-RAW"
            self.imap_conn.literal = b"(%s)" % criterion.replace("gmail:", "").encode(
                "utf-8"
            )
        else:
            search_command = "(%s)" % criterion
        try:
            self.logger.info("IMAP search: '%s'", criterion)
            status, data = self.imap_conn.search(search_encoding, search_command)
        except Exception as err:  # pylint: disable=broad-except
            self.logger.warning(
                "Email search returned: %s (%s)", str(err), search_command
            )
            return False
        if status == "OK":
            mail_id_data = bytes.decode(data[0])
            mail_ids = mail_id_data.split() if len(mail_id_data) > 0 else []
            if limit and len(mail_ids) + result["message_count"] > limit:
                return True
            if len(mail_ids) > 0:
                result["message_count"] += len(mail_ids)
                result["ids"].extend(mail_ids)
                for mail_id in mail_ids:
                    mail_uid, message = self._fetch_uid_and_message(mail_id, actions)
                    if mail_uid is None or mail_uid in result["uids"].keys():
                        continue
                    message["uid"] = mail_uid
                    result["uids"][mail_uid] = message
            return True
        else:
            self.logger.warning("Search result not OK: %s", status)
            return False

    def _get_mail_uids_and_dicts(self, mail):
        mail_uids = []
        mail_dicts = []
        if isinstance(mail, list):
            mail_uids = [m["uid"] for m in mail]
            mail_dict = None
            mail_dicts = mail
        else:
            try:
                mail_uid, mail_dict = mail
                mail_uids = [mail_uid]
                mail_dicts = [mail_dict]
            except ValueError:
                mail_uid = mail["uid"]
                mail_dict = mail
                mail_uids = [mail_uid]
                mail_dicts = [mail]
        return mail_uids, mail_dicts

    def _command_mail_uid(self, command, uid, m_op="", m_param="", target_folder=None):
        if target_folder:
            action_status, _ = self.imap_conn.uid(command, uid, target_folder)
            self.logger.debug(
                "%s %s %s RESULT %s", command, uid, target_folder, action_status
            )
        else:
            action_status, _ = self.imap_conn.uid(command, uid, m_op, m_param)
            self.logger.info(
                "%s %s %s %s RESULT %s",
                command,
                uid,
                m_op,
                m_param,
                action_status,
            )
        return action_status

    def _perform_actions(  # noqa: C901 pylint: disable=too-many-branches
        self,
        actions: list,
        mail: Union[dict, list],
        labels: str = None,
        target_folder: str = None,
        overwrite: bool = False,
        prefix: str = "",
    ):
        action_status = True
        mail_uids, mail_dicts = self._get_mail_uids_and_dicts(mail)
        store_params = {
            Action.glabel_add: ["+X-GM-LABELS", f"({labels})"],
            Action.glabel_remove: ["-X-GM-LABELS", f"({labels})"],
            Action.msg_delete: ["+FLAGS", FLAG_DELETED],
            Action.msg_flag: ["+FLAGS", FLAG_FLAGGED],
            Action.msg_unflag: ["-FLAGS", FLAG_FLAGGED],
            Action.msg_read: ["+FLAGS", FLAG_SEEN],
            Action.msg_unread: ["-FLAGS", FLAG_SEEN],
            Action.msg_trash: ["+FLAGS", FLAG_TRASH],
        }
        for act in actions:
            action = to_action(act)
            if action == Action.msg_list:
                continue
            if action in store_params.keys():
                m_op, m_param = store_params[action]
                for uid in mail_uids:
                    action_status = self._command_mail_uid("STORE", uid, m_op, m_param)
            elif action == Action.msg_copy:
                for uid in mail_uids:
                    action_status = self._command_mail_uid(
                        "COPY", uid, target_folder=target_folder
                    )
            elif action == Action.msg_move:
                m_op, m_param = store_params[Action.msg_delete]
                for uid in mail_uids:
                    if " " in target_folder:
                        target_folder = f'"{target_folder}"'
                    action_status = self._command_mail_uid(
                        "COPY", uid, target_folder=target_folder
                    )
                    action_status = self._command_mail_uid("STORE", uid, m_op, m_param)
            elif action == Action.msg_save and len(mail_dicts) > 0:
                if target_folder is None:
                    target_folder = os.path.expanduser("~")
                for mail_dict in mail_dicts:
                    action_status = self._save_eml_file(
                        mail_dict, target_folder, overwrite, prefix
                    )
            elif action == Action.msg_attachment_save and len(mail_dicts) > 0:
                if target_folder is None:
                    target_folder = os.path.expanduser("~")
                for mail_dict in mail_dicts:
                    self._save_attachment(mail_dict, target_folder, overwrite, prefix)
            else:
                # TODO: mypy should handle enum exhaustivity validation
                raise ValueError(
                    f"Unsupported email action or insufficient input data: {action}"
                )
        return action_status

    @imap_connection
    def delete_message(
        self,
        criterion: Union[str, dict] = None,
        source_folder: str = None,
    ) -> bool:
        """Delete single message from server based on criterion.

        :param criterion: filter messages based on this search, can also be a
         message dictionary
        :param source_folder: defaults to already selected folder, but can be
         set to delete message in a specific folder
        :return: True if success, False if not

        If criterion does not return exactly 1 message then delete is not done.

        Example:

        .. code-block:: robotframework

            Delete Message  SUBJECT \"Greetings RPA developer\"
        """
        selected_folder = source_folder or self.selected_folder
        if isinstance(criterion, dict):
            readonly = False
            self.select_folder(selected_folder, readonly)
            status = self._perform_actions(
                [Action.msg_delete],
                criterion,
            )
            result = {
                "actions_done": 1 if status else 0,
                "message_count": 1,
                "ids": [criterion["uid"]],
                "uids": {criterion["uid"]: criterion},
            }
            self.imap_conn.expunge()
        else:
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_delete],
                limit=1,
                source_folder=selected_folder,
            )
            self.imap_conn.expunge()
        return result["actions_done"] == 1

    @imap_connection
    def delete_messages(
        self,
        criterion: Union[str, list] = None,
        limit: int = None,
        source_folder: str = None,
    ) -> bool:
        """Delete messages from server based on criterion.

        :param criterion: filter messages based on this search, can also be a
         list of message dictionaries
        :param limit: maximum number of message to delete
        :param source_folder: defaults to already selected folder, but can be
         set to delete message in a specific folder
        :return: True if success, False if not

        Example:

        .. code-block:: robotframework

            Delete Messages  SUBJECT Greetings
        """
        selected_folder = source_folder or self.selected_folder
        if isinstance(criterion, list):
            readonly = False
            self.select_folder(selected_folder, readonly)
            status = self._perform_actions(
                [Action.msg_delete],
                criterion,
            )
            self.imap_conn.expunge()
            return status
        else:
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_delete],
                limit=limit,
                source_folder=selected_folder,
            )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @imap_connection
    def save_messages(
        self,
        criterion: Optional[Union[str, dict, list]] = None,
        target_folder: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> bool:
        # pylint: disable=C0301
        """Save messages based on criteria and store them to target folder
        with attachment files.

        Does not save message if `target_folder` is not given.

        :param criterion: filter messages based on this, defaults to ""
        :param target_folder: path to folder where message are saved, defaults to None
        :param prefix: optional filename prefix added to the attachments, empty by
            default
        :return: True if succeeded, False otherwise

        Example:

        .. code-block:: robotframework

            Save Messages  SUBJECT Important message  target_folder=${USERDIR}${/}messages
        """  # noqa: E501
        prefix = prefix or ""
        if target_folder is None:
            target_folder = os.path.expanduser("~")
        if isinstance(criterion, (dict, list)):
            status = self._perform_actions(
                [Action.msg_save], criterion, target_folder=target_folder, prefix=prefix
            )
            return status
        else:
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_save],
                target_folder=target_folder,
                prefix=prefix,
            )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @imap_connection
    def list_messages(
        self, criterion: str = "", source_folder: str = None, readonly: bool = True
    ) -> Any:
        """Return list of messages matching criterion.

        :param criterion: list emails matching this, defaults to ""
        :param source_folder: list messages from this folder
        :param readonly: set False if you want to mark matching messages as read
        :return: list of messages

        *Note.* listing messages without `source_folder` might take a long time

        Example:

        .. code-block:: robotframework

            @{emails}  List Messages  SUBJECT "rpa task"
            FOR  ${email}  IN  @{EMAILS}
                Log  ${email}[Subject]
                Log  ${email}[From]
                Log  ${email}[Date]
                Log  ${email}[Delivered-To]
                Log  ${email}[Received]
                Log  ${email}[Has-Attachments]
                Log  ${email}[uid]
            END
        """
        result = self._do_actions_on_messages(
            criterion,
            source_folder=source_folder,
            actions=[Action.msg_list],
            readonly=readonly,
        )
        values = result["uids"].values()
        converted = []
        for v in values:
            if v:
                converted.append(
                    {
                        str(key): value
                        if isinstance(value, (str, bool, int, Message))
                        else str(value)
                        for key, value in v.items()
                    }
                )
        return converted

    @imap_connection
    def save_attachments(
        self,
        criterion: str = "",
        target_folder: Optional[str] = None,
        overwrite: bool = False,
        prefix: Optional[str] = None,
    ) -> List[str]:
        # pylint: disable=C0301
        """Save mail attachments of emails matching criterion on the local disk.

        :param criterion: attachments are saved for mails matching this, defaults to ""
        :param target_folder: local folder for saving attachments to (needs to exist),
            defaults to user's home directory if None
        :param overwrite: overwrite existing file if True, defaults to False
        :param prefix: optional filename prefix added to the attachments, empty by
            default
        :return: list of saved attachments (absolute file paths) of all emails

        Example:

        .. code-block:: robotframework

            ${attachments} =    Save Attachments    SUBJECT "rpa task"
            ...    target_folder=${CURDIR}${/}messages  overwrite=${True}
            FOR  ${file}  IN  @{attachments}
                OperatingSystem.File Should Exist  ${file}
            END
        """  # noqa: E501
        attachments_saved = []
        prefix = prefix or ""
        messages = self.list_messages(criterion)
        for msg in messages:
            attachments_saved.extend(
                self.save_attachment(msg, target_folder, overwrite, prefix)
            )
        return attachments_saved

    def save_attachment(
        self,
        message: Union[dict, Message],
        target_folder: Optional[str],
        overwrite: bool,
        prefix: Optional[str] = None,
    ) -> List[str]:
        # pylint: disable=C0301
        """Save mail attachment of a single given email on the local disk.

        :param message: message item
        :param target_folder: local folder for saving attachments to (needs to exist),
            defaults to user's home directory if None
        :param overwrite: overwrite existing file if True, defaults to False
        :param prefix: optional filename prefix added to the attachments, empty by
            default
        :return: list of saved attachments (list of absolute filepaths) in one email

        Example:

        .. code-block:: robotframework

            @{emails} =    List Messages    ALL
            FOR    ${email}    IN    @{emails}
                IF    ${email}[Has-Attachments]
                    Log To Console    Saving attachment for: ${email}[Subject]
                    ${attachments} =    Save Attachment
                    ...    ${email}
                    ...    target_folder=${CURDIR}
                    ...    overwrite=${True}
                    Log To Console    Saved attachments: ${attachments}
                END
            END
        """  # noqa: E501
        prefix = prefix or ""
        if target_folder is None:
            target_folder = os.path.expanduser("~")
        return self._save_attachment(message, target_folder, overwrite, prefix)

    def _save_attachment(self, message, target_folder, overwrite, prefix) -> List[str]:
        attachments_saved = []
        msg = message["Message"] if isinstance(message, dict) else message

        for part in msg.walk():
            content_maintype = part.get_content_maintype()
            filename = None
            if content_maintype != "multipart":
                filename = get_part_filename(part)
            if not filename:
                continue

            filepath = Path(target_folder) / Path(f"{prefix}{filename}").name
            if not overwrite:
                filepath = counter_duplicate_path(filepath)
            self.logger.info("Attachment filepath: %r", filepath)
            payload = part.get_payload(decode=True)
            if payload:
                self.logger.info(
                    "Saving attachment: %s",
                    filename,
                )
                with open(filepath, "wb") as f:
                    f.write(payload)
                    attachments_saved.append(str(filepath))
            else:
                self.logger.info(
                    "Attachment %r did not have payload to write",
                    filename,
                )
        return attachments_saved

    def _save_eml_file(self, message, target_folder, overwrite, prefix):
        save_status = True
        emlfile = f"{prefix}{message['uid']}.eml"
        full_emlfile_path = Path(os.path.join(target_folder, emlfile))
        if not full_emlfile_path.exists() or overwrite:
            with open(full_emlfile_path, "wb") as f:
                if "bytes" in message.keys():
                    f.write(message["bytes"])
                elif "Message" in message:
                    f.write(message["Message"].as_bytes())
                else:
                    save_status = False
                    self.logger.error(
                        "Save failed. Unable to get message byte content."
                    )
        elif full_emlfile_path.exists() and not overwrite:
            self.logger.warning("Did not overwrite file: %s", full_emlfile_path)
        return save_status

    @imap_connection
    def wait_for_message(
        self,
        criterion: str = "",
        timeout: float = 5.0,
        interval: float = 1.0,
        readonly: bool = True,
    ) -> Any:
        """Wait for email matching `criterion` to arrive into mailbox.

        :param criterion: message filter to wait for, defaults to ""
        :param timeout: total time in seconds to wait for email, defaults to 5.0
        :param interval: time in seconds for new check, defaults to 1.0
        :param readonly: set False if you want to mark matching messages as read
        :return: list of messages

        Example:

        .. code-block:: robotframework

            @{emails}  Wait For Message  SUBJECT \"rpa task\"  timeout=300  interval=10
        """
        self._validate_criterion(criterion)
        end_time = time.time() + float(timeout)
        while time.time() < end_time:
            self.imap_conn.select("inbox", readonly)
            result = self._do_actions_on_messages(
                criterion,
                actions=[Action.msg_list],
                source_folder="INBOX",
                readonly=readonly,
            )
            if result["message_count"] > 0:
                return result["uids"].values()
            time.sleep(interval)
        return []

    def _parse_folders(self, folders):
        if folders and len(folders) == 1 and folders[0] is None:
            return []
        parsed_folders = []
        folder_regex = r'\((?P<flags>.*)\)\s"(?P<delimiter>.*)"\s"?(?P<name>[^"]*)"?'
        for f in folders:
            match = re.search(folder_regex, bytes.decode(f))
            if not match:
                self.logger.warning("Cannot parse folder name %s", bytes.decode(f))
                continue
            parsed_folders.append(match.groupdict())
        return parsed_folders

    @imap_connection
    def get_folder_list(self, subdirectory: str = None, pattern: str = None) -> list:
        """Get list of folders on the server

        :param subdirectory: list subdirectories for this folder
        :param pattern: list folders matching this pattern
        :return: list of folders

        Example:

        .. code-block:: robotframework

            @{folders}  Get Folder List
            @{folders}  Get Folder List  pattern=important
            @{folders}  Get Folder List  subdirectory=sub
        """
        kwparams = {}
        if subdirectory:
            kwparams["directory"] = f'"{subdirectory}"'
        if pattern:
            kwparams["pattern"] = pattern

        status_code, folders = self.imap_conn.list(**kwparams)

        if status_code == "OK":
            parsed_folders = self._parse_folders(folders)
            return [
                f
                for f in parsed_folders
                if subdirectory is None or subdirectory == f["name"]
            ]
        else:
            return []

    @imap_connection
    def select_folder(self, folder_name: str = "INBOX", readonly: bool = False) -> int:
        """Select folder by name

        :param folder_name: name of the folder to select
        :param readonly: if set to True then message flags are not modified
        :return: message count in the selected folder

        Returns number of messages in the folder or
        exception if folder does not exist on the server.

        Example:

        .. code-block:: robotframework

            Select Folder   subfolder
        """
        status_code, data = self.imap_conn.select(
            mailbox=f'"{folder_name}"', readonly=readonly
        )
        if status_code == "OK":
            message_count = bytes.decode(data[0])
            self.selected_folder = folder_name
            return int(message_count)
        else:
            raise ValueError("Folder '%s' does not exist on the server" % folder_name)

    @imap_connection
    def rename_folder(
        self, oldname: str = None, newname: str = None, suppress_error: bool = False
    ) -> bool:
        """Rename email folder

        :param oldname: current folder name
        :param newname: new name for the folder
        :param suppress_error: to silence warning message, defaults to False
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Rename Folder   subfolder   filtered
        """
        if oldname is None or newname is None:
            raise KeyError(
                "Both 'oldname' and 'newname' and required for rename folder"
            )
        self.logger.info("Rename folder '%s' to '%s'", oldname, newname)
        status_code, data = self.imap_conn.rename(oldname, newname)

        if status_code == "OK":
            return True
        else:
            if suppress_error is False:
                self.logger.warning(
                    "Folder rename failed with message: '%s'",
                    bytes.decode(data[0]),
                )
            return False

    @imap_connection
    def delete_folder(self, folder_name: str = None) -> bool:
        """Delete email folder

        :param folder_name: current folder name
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Delete Folder   filtered
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for delete folder")
        self.logger.info("Delete folder '%s'", folder_name)
        status_code, data = self.imap_conn.delete(folder_name)
        if status_code == "OK":
            return True
        else:
            self.logger.warning(
                "Delete folder '%s' response status: '%s %s'",
                folder_name,
                status_code,
                bytes.decode(data[0]),
            )
            return False

    @imap_connection
    def create_folder(self, folder_name: str = None) -> bool:
        """Create email folder

        :param folder_name: name for the new folder
        :return: True if operation was successful, False if not

        Example:

        .. code-block:: robotframework

            Create Folder   filtered
        """
        if folder_name is None:
            raise KeyError("'folder_name' is required for create folder")
        self.logger.info("Create folder '%s'", folder_name)
        status_code, data = self.imap_conn.create(folder_name)
        if status_code == "OK":
            return True
        else:
            self.logger.warning(
                "Create folder '%s' response status: '%s %s'",
                folder_name,
                status_code,
                bytes.decode(data[0]),
            )
            return False

    @imap_connection
    def flag_messages(
        self, criterion: Union[str, dict] = None, unflag: bool = False
    ) -> Any:
        """Mark messages as `flagged`

        :param criterion: mark messages matching criterion
        :param unflag: to mark messages as not `flagged`
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${flagged}  ${oftotal}    Flag Messages   SUBJECT rpa
            ${unflagged}  ${oftotal}  Flag Messages   SUBJECT rpa  unflag=True
        """
        action = Action.msg_unflag if unflag else Action.msg_flag
        if isinstance(criterion, dict):
            status = self._perform_actions(mail=criterion, actions=[action])
            return status
        else:
            result = self._do_actions_on_messages(criterion, actions=[action])
            return (
                result["actions_done"] > 0
                and result["actions_done"] == result["message_count"]
            )

    @imap_connection
    def unflag_messages(self, criterion: Union[str, dict] = None) -> Any:
        """Mark messages as not `flagged`

        :param criterion: mark messages matching criterion
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${unflagged}  ${oftotal}  Unflag Messages   SUBJECT rpa
        """
        return self.flag_messages(criterion, unflag=True)

    @imap_connection
    def mark_as_read(
        self, criterion: Union[str, dict] = None, unread: bool = False
    ) -> Any:
        """Mark messages as `read`

        :param criterion: mark messages matching criterion
        :param unread: to mark messages as not `read`
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${read}  ${oftotal}  Mark As Read   SUBJECT rpa
        """
        action = Action.msg_unread if unread else Action.msg_read
        if isinstance(criterion, dict):
            status = self._perform_actions(mail=criterion, actions=[action])
            return status
        else:
            result = self._do_actions_on_messages(criterion, actions=[action])
            return (
                result["actions_done"] > 0
                and result["actions_done"] == result["message_count"]
            )

    @imap_connection
    def mark_as_unread(self, criterion: Union[str, dict] = None) -> Any:
        """Mark messages as not `read`

        :param criterion: mark messages matching criterion
        :return: successful operations (int), matching messages (int)

        Example:

        .. code-block:: robotframework

            ${unread}  ${oftotal}  Mark As Unread   SUBJECT rpa
        """
        return self.mark_as_read(criterion, unread=True)

    def _validate_criterion(self, criterion: str) -> bool:
        if criterion is None or len(criterion) < 1:
            raise KeyError("Criterion is required parameter")

    def _fetch_uid_and_message(self, mail_id, actions):
        message_dict = None
        _, data = self.imap_conn.fetch(mail_id, "(UID RFC822)")
        pattern_uid = re.compile(r".*UID (\d+) RFC822")
        decoded_data = bytes.decode(data[0][0]) if data[0] else None
        self.logger.debug("message identification: %s", decoded_data)
        match_result = pattern_uid.match(decoded_data) if decoded_data else None
        uid = match_result.group(1) if match_result else None
        if uid:
            message_dict = self._fetch_message_dict(mail_id, data, actions)

        key_to_change = None
        message_id_to_add = None
        if message_dict:
            for key, val in message_dict.items():
                if key.lower() == "message-id":
                    key_to_change = key
                    message_id_to_add = (
                        str(val).replace("<", "").replace(">", "").strip()
                    )
            if key_to_change:
                message_dict["Message-ID"] = message_id_to_add
                if key_to_change != "Message-ID":
                    del message_dict[key_to_change]
        return uid, message_dict

    def _fetch_message_dict(self, mail_id, data, actions):
        # _, data = self.imap_conn.fetch(mail_id, "(RFC822)")
        message = message_from_bytes(data[0][1])
        message_dict = {"Mail-Id": mail_id, "Message": message, "Body": ""}
        if Action.msg_save in actions:
            message_dict["bytes"] = data[0][1]
        for k, v in message.items():
            msg_item = decode_header(v)
            message_dict[k] = make_header(msg_item)
        (
            message_dict["Body"],
            has_attachments,
        ) = self.get_decoded_email_body(message)
        # SET DEFAULT VALUES FOR KEYS
        if "Delivered-To" not in message_dict.keys():
            message_dict["Delivered-To"] = ""
        message_dict["Has-Attachments"] = has_attachments
        return message_dict

    @imap_connection
    def move_messages(
        self,
        criterion: Union[str, dict] = None,
        target_folder: str = None,
        source_folder: str = None,
    ) -> bool:
        """Move messages from source folder to target folder

        :param criterion: move messages matching criterion
        :param source_folder: location of the messages, default `INBOX`
        :param target_folder: where messages should be move into
        :return: True if all move operations succeeded, False if not

        Example:

        .. code-block:: robotframework

            ${result}=    Move Messages
            ...    criterion=SUBJECT "order confirmation 32"
            ...    target_folder=yyy

            ${result}=    Move Messages
            ...    criterion=ALL
            ...    source_folder=yyy
            ...    target_folder=XXX
        """
        selected_folder = source_folder or self.selected_folder
        if target_folder is None or len(target_folder) == 0:
            raise KeyError("Can't move messages without target_folder")
        if isinstance(criterion, dict):
            readonly = False
            self.select_folder(selected_folder, readonly)
            status = self._perform_actions(
                mail=criterion,
                actions=[Action.msg_move],
                target_folder=target_folder,
            )
            result = {
                "actions_done": 1 if status else 0,
                "message_count": 1,
                "ids": [criterion["uid"]],
                "uids": {criterion["uid"]: criterion},
            }
        else:
            result = self._do_actions_on_messages(
                criterion=criterion,
                actions=[Action.msg_move],
                source_folder=source_folder,
                target_folder=target_folder,
            )
        if result["actions_done"] > 0:
            self.imap_conn.expunge()
        action_mismatch = result["actions_done"] != result["message_count"]
        if action_mismatch:
            self.logger.warning(
                "Criterion matched %s messages, but actions done to %s messages",
                result["message_count"],
                result["actions_done"],
            )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @imap_connection
    def move_messages_by_ids(
        self,
        message_ids: Union[str, List],
        target_folder: str,
        source_folder: str,
        use_gmail_search: bool = False,
    ) -> bool:
        """Move message by their Message-ID's from source folder to target folder

        :param message_ids: one Message-ID as string or list of Message-IDs
        :param source_folder: location of the messages, default `INBOX`
        :param target_folder: where messages should be move into
        :param use_gmail_search: set to True to use `Rfc822msgid` search, default
            is `HEADER Message-ID` search
        :return: True if all move operations succeeded, False if not
        """
        if not message_ids:
            self.logger.warning("No message ids given for Move Messages By IDs")
            return False
        idlist = message_ids if isinstance(message_ids, list) else [message_ids]
        results = []
        for mid in idlist:
            if use_gmail_search:
                move_criterion = f"gmail:Rfc822msgid:{mid}"
            else:
                move_criterion = f"HEADER Message-ID {mid}"
            result = self.move_messages(
                criterion=move_criterion,
                source_folder=source_folder,
                target_folder=target_folder,
            )
            if not result:
                self.logger.warning("Moving Message-ID '%s' failed", mid)
            results.append(result)
        return all(results)

    @imap_connection
    def _modify_gmail_labels(
        self,
        labels: str,
        criterion: Union[str, dict],
        action: bool = True,
        source_folder: str = None,
    ) -> bool:
        if isinstance(criterion, dict):
            selected_folder = source_folder or self.selected_folder
            self.select_folder(selected_folder)
            status = self._perform_actions(
                mail=criterion,
                actions=[action],
                labels=labels,
            )
            return status
        else:
            result = self._do_actions_on_messages(
                criterion=criterion,
                actions=[action],
                labels=labels,
                source_folder=source_folder,
            )

            return (
                result["actions_done"] > 0
                and result["actions_done"] == result["message_count"]
            )

    @imap_connection
    def add_gmail_labels(self, labels, criterion, source_folder: str = None) -> bool:
        """Add GMail labels to messages matching criterion and if given,
        source folder

        :param labels: comma separated list of labels to add
        :param criterion: label messages matching criterion
        :param source_folder: look for messages in this folder, default all folders
        :return: status of the operation

        Example:

        .. code-block:: robotframework

            Add Gmail Labels  customer1   SUBJECT "order confirmation"
            Add Gmail Labels  wip         SUBJECT "order confirmation"   customerfolder
        """
        return self._modify_gmail_labels(
            labels, criterion, Action.glabel_add, source_folder
        )

    @imap_connection
    def remove_gmail_labels(self, labels, criterion, source_folder: str = None) -> bool:
        """Remove GMail labels to messages matching criterion and if given,
        source folder

        :param labels: comma separated list of labels to remove
        :param criterion: unlabel messages matching criterion
        :param source_folder: look for messages in this folder, default all folders
        :return: status of the operation

        Example:

        .. code-block:: robotframework

            Remove Gmail Labels  wip  SUBJECT "order confirmation"
            Remove Gmail Labels  wip  SUBJECT "order confirmation"  customerfolder
        """
        return self._modify_gmail_labels(
            labels, criterion, Action.glabel_remove, source_folder
        )

    @imap_connection
    def do_message_actions(
        self,
        criterion: str = "",
        actions: list = None,
        source_folder: str = None,
        target_folder: str = None,
        labels: str = None,
        limit: int = None,
        overwrite: bool = False,
        prefix: str = None,
    ) -> Any:
        """Do actions to messages matching criterion and if given,
        source folder

        Actions can be:

        - msg_copy
        - msg_delete
        - msg_flag
        - msg_unflag
        - msg_read
        - msg_unread
        - msg_save
        - msg_attachment_save
        - glabel_add
        - glabel_remove

        Result object contains following attributes:

        - actions_done, number of messages on which action was performed
        - message_count, number of messages matching criterion
        - ids, message ids matching criterion
        - uids, dictionary of message uids and message content

        :param criterion: perform actions on messages matching this
        :param actions: list of actions to perform on matching messages
        :param source_folder: look for messages in this folder, default all folders
        :param target_folder: can be file path or email folder
         (for example action: msg_copy)
        :param labels: comma separated list of labels (for example action: glabel_add)
        :param limit:  maximum number of messages (for example action: msg_delete)
        :param overwrite: to control if file should overwrite
         (for example action: msg_attachment_save)
        :param prefix: prefix to be added into filename (for example: msg_save)
        :return: result object

        Example:

        .. code-block:: robotframework

            ${actions}=   Create List  msg_unflag  msg_read  msg_save  msg_attachment_save
            Do Message Actions    SUBJECT "Order confirmation"
            ...                   ${actions}
            ...                   source_folder=XXX
            ...                   target_folder=${CURDIR}
            ...                   overwrite=True
        """  # noqa: E501

        parsed_actions = [to_action(act) for act in actions]
        result = self._do_actions_on_messages(
            criterion=criterion,
            actions=parsed_actions,
            labels=labels,
            source_folder=source_folder,
            target_folder=target_folder,
            limit=limit,
            overwrite=overwrite,
            prefix=prefix,
        )
        return (
            result["actions_done"] > 0
            and result["actions_done"] == result["message_count"]
        )

    @staticmethod
    def _ensure_path_object(source: FilePath) -> Path:
        if not isinstance(source, Path):
            source = Path(source)
        return source.expanduser().resolve()

    def email_to_document(
        self, input_source: Union[FilePath, BinaryIO, bytes], output_path: FilePath
    ):
        """Convert a raw e-mail into a Word document.

        This keyword extracts the HTML (or Text) content from the passed input e-mail
        and saves it into docx format at the provided output path.

        :param input_source: Path, bytes or file-like object with the input raw e-mail
            content
        :param output_path: Where to save the output docx file

        Example:

        **Robot Framework**

        .. code-block:: robotframework

            Convert email to docx
                ${mail_file} =     Get Work Item File    mail.eml
                Email To Document    ${mail_file}    ${OUTPUT_DIR}${/}mail.docx

        **Python**

        .. code-block:: python

            from pathlib import Path
            from RPA.Email.ImapSmtp import ImapSmtp
            from RPA.Robocorp.WorkItems import WorkItems

            lib_work = WorkItems()
            lib_mail = ImapSmtp()

            def convert_email_to_docx():
                lib_work.get_input_work_item()
                mail_file = lib_work.get_work_item_file("mail.eml")
                lib_mail.email_to_document(mail_file, Path("./output") / "mail.docx")

            convert_email_to_docx()
        """

        if hasattr(input_source, "read"):
            self.logger.info("Reading raw e-mail bytes from the provided source object")
            data = input_source.read()
        elif isinstance(input_source, bytes):
            self.logger.info("Using the provided source bytes as raw e-mail content")
            data = input_source
        else:
            input_source = self._ensure_path_object(input_source)
            self.logger.info("Reading raw e-mail bytes from: %s", input_source)
            data = input_source.read_bytes()
        assert isinstance(
            data, bytes
        ), "bytes expected for e-mail parsing, got %s" % type(data)

        self.logger.info("Getting the html/text from the raw e-mail")
        message = message_from_bytes(data)
        body, _ = self.get_decoded_email_body(message, html_first=True)

        h2d_parser = HtmlToDocx()
        self.logger.debug("Converting html/text content:\n%s", body)
        docx = h2d_parser.parse_html_string(body)
        output_path = self._ensure_path_object(output_path)
        self.logger.info("Writing converted document into: %s", output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        docx.save(output_path)



try:
    import RPA.recognition as _unused

    del _unused
    HAS_RECOGNITION = True
except ImportError:
    HAS_RECOGNITION = False

from typing import Optional


class ElementNotFound(ValueError):
    """No matching elements were found."""


class MultipleElementsFound(ValueError):
    """Multiple matching elements were found, but only one was expected."""


class TimeoutException(ValueError):
    """Timeout reached while waiting for condition."""


class LibraryContext:
    """Shared context for all keyword libraries."""

    def __init__(self, ctx):
        self.ctx = ctx

    @property
    def logger(self):
        return self.ctx.logger

    @property
    def buffer(self):
        return self.ctx.buffer

    @property
    def locators_path(self) -> Optional[str]:
        return self.ctx.locators_path
